from __future__ import division, print_function

import os
import tempfile
import warnings

import matplotlib.pyplot as plt
import numpy as np
from numpy.testing import assert_array_equal
import pytest

from astropy.utils.data import get_pkg_data_filename
from astropy.io.fits.verify import VerifyWarning
from astropy.coordinates import SkyCoord
from astropy.io import fits
from astropy import wcs
from astropy.io.fits.card import UNDEFINED
import astropy.units as u
from astropy.utils.exceptions import AstropyWarning

from ..targetpixelfile import KeplerTargetPixelFile, TargetPixelFileFactory
from ..targetpixelfile import TessTargetPixelFile, TargetPixelFile
from ..lightcurve import TessLightCurve
from ..utils import LightkurveWarning, LightkurveDeprecationWarning
from ..io import read
from ..search import search_tesscut

from .test_synthetic_data import filename_synthetic_flat

filename_tpf_all_zeros = get_pkg_data_filename("data/test-tpf-all-zeros.fits")
filename_tpf_one_center = get_pkg_data_filename("data/test-tpf-non-zero-center.fits")
filename_tess = get_pkg_data_filename("data/tess25155310-s01-first-cadences.fits.gz")

TABBY_Q8 = ("https://archive.stsci.edu/missions/kepler/lightcurves"
            "/0084/008462852/kplr008462852-2011073133259_llc.fits")
TABBY_TPF = ("https://archive.stsci.edu/missions/kepler/target_pixel_files"
             "/0084/008462852/kplr008462852-2011073133259_lpd-targ.fits.gz")
TESS_SIM = ("https://archive.stsci.edu/missions/tess/ete-6/tid/00/000"
            "/004/176/tess2019128220341-0000000417699452-0016-s_tp.fits")
asteroid_TPF = get_pkg_data_filename("data/asteroid_test.fits")


@pytest.mark.remote_data
def test_load_bad_file():
    '''Test if a light curve can be opened without exception.'''
    with pytest.raises(ValueError) as exc:
        KeplerTargetPixelFile(TABBY_Q8)
    assert('is this a target pixel file?' in exc.value.args[0])
    with pytest.raises(ValueError) as exc:
        TessTargetPixelFile(TABBY_Q8)
    assert('is this a target pixel file?' in exc.value.args[0])


def test_tpf_shapes():
    """Are the data array shapes of the TargetPixelFile object consistent?"""
    with warnings.catch_warnings():
        # Ignore the "TELESCOP is not equal to TESS" warning
        warnings.simplefilter("ignore", LightkurveWarning)
        tpfs = [KeplerTargetPixelFile(filename_tpf_all_zeros),
                TessTargetPixelFile(filename_tpf_all_zeros)]
    for tpf in tpfs:
        assert tpf.quality_mask.shape == tpf.hdu[1].data['TIME'].shape
        assert tpf.flux.shape == tpf.flux_err.shape


def test_tpf_math():
    """Can you add, subtract, multiply and divide?"""
    with warnings.catch_warnings():
        # Ignore the "TELESCOP is not equal to TESS" warning
        warnings.simplefilter("ignore", LightkurveWarning)
        tpfs = [KeplerTargetPixelFile(filename_tpf_all_zeros),
                TessTargetPixelFile(filename_tpf_all_zeros)]

        # These should work
        for tpf in tpfs:
            for other in [1, np.ones(tpf.flux.shape[1:]), np.ones(tpf.shape)]:
                tpf + other
                tpf - other
                tpf * other
                tpf / other

                tpf += other
                tpf -= other
                tpf *= other
                tpf /= other

        # These should fail with a value error because their shape is wrong.
        for tpf in tpfs:
            for other in [np.asarray([1, 2]), np.arange(len(tpf.time) - 1),
                          np.ones([100, 1]), np.ones([1, 2, 3])]:
                with pytest.raises(ValueError):
                    tpf + other

        # Check the values are correct
        assert np.all(((tpf.flux.value + 2) == (tpf + 2).flux.value)[np.isfinite(tpf.flux)])
        assert np.all(((tpf.flux.value - 2) == (tpf - 2).flux.value)[np.isfinite(tpf.flux)])
        assert np.all(((tpf.flux.value * 2) == (tpf * 2).flux.value)[np.isfinite(tpf.flux)])
        assert np.all(((tpf.flux.value / 2) == (tpf / 2).flux.value)[np.isfinite(tpf.flux)])
        assert np.all(((tpf.flux_err.value * 2) == (tpf * 2).flux_err.value)[np.isfinite(tpf.flux)])
        assert np.all(((tpf.flux_err.value / 2) == (tpf / 2).flux_err.value)[np.isfinite(tpf.flux)])


def test_tpf_plot():
    """Sanity check to verify that tpf plotting works"""
    with warnings.catch_warnings():
        # Ignore the "TELESCOP is not equal to TESS" warning
        warnings.simplefilter("ignore", LightkurveWarning)
        tpfs = [KeplerTargetPixelFile(filename_tpf_one_center),
                TessTargetPixelFile(filename_tpf_one_center)]
    for tpf in tpfs:
        tpf.plot()
        tpf.plot(aperture_mask=tpf.pipeline_mask)
        tpf.plot(aperture_mask='all')
        tpf.plot(frame=3)
        with pytest.raises(ValueError):
            tpf.plot(frame=999999)
        tpf.plot(cadenceno=125250)
        with pytest.raises(ValueError):
            tpf.plot(cadenceno=999)
        tpf.plot(bkg=True)
        tpf.plot(scale="sqrt")
        tpf.plot(scale="log")
        with pytest.raises(ValueError):
            tpf.plot(scale="blabla")
        tpf.plot(column='FLUX')
        tpf.plot(column='FLUX_ERR')
        tpf.plot(column='FLUX_BKG')
        tpf.plot(column='FLUX_BKG_ERR')
        tpf.plot(column='RAW_CNTS')
        tpf.plot(column='COSMIC_RAYS')
        with pytest.raises(ValueError):
            tpf.plot(column='not a column')

        plt.close('all')


def test_tpf_zeros():
    """Does the LightCurve of a zero-flux TPF make sense?"""
    tpf = KeplerTargetPixelFile(filename_tpf_all_zeros, quality_bitmask=None)
    with warnings.catch_warnings():
        # Ignore "LightCurve contains NaN times" warnings triggered by the liberal mask
        warnings.simplefilter("ignore", LightkurveWarning)
        lc = tpf.to_lightcurve()
    # If you don't mask out bad data, time contains NaNs
    assert np.any(lc.time.value != tpf.time)  # Using the property that NaN does not equal NaN
    # When you do mask out bad data everything should work.
    assert (tpf.time.value == 0).any()
    tpf = KeplerTargetPixelFile(filename_tpf_all_zeros, quality_bitmask='hard')
    lc = tpf.to_lightcurve(aperture_mask="all")
    assert len(lc.time) == len(lc.flux)
    assert np.all(lc.time == tpf.time)
    assert np.all(np.isnan(lc.flux)) # we expect all NaNs because of #874
    # The default QUALITY bitmask should have removed all NaNs in the TIME
    assert ~np.any(np.isnan(tpf.time.value))

@pytest.mark.parametrize("centroid_method", [("moments"), ("quadratic")])
def test_tpf_ones(centroid_method):
    """Does the LightCurve of a one-flux TPF make sense?"""
    with warnings.catch_warnings():
        # Ignore the "TELESCOP is not equal to TESS" warning
        warnings.simplefilter("ignore", LightkurveWarning)
        tpfs = [KeplerTargetPixelFile(filename_tpf_one_center),
                TessTargetPixelFile(filename_tpf_one_center)]
    for tpf in tpfs:
        lc = tpf.to_lightcurve(aperture_mask='all', centroid_method=centroid_method)
        assert np.all(lc.flux.value == 1)
        assert np.all((lc.centroid_col.value < tpf.column+tpf.shape[1]).all()
                      * (lc.centroid_col.value > tpf.column).all())
        assert np.all((lc.centroid_row.value < tpf.row+tpf.shape[2]).all()
                      * (lc.centroid_row.value > tpf.row).all())


@pytest.mark.parametrize("quality_bitmask,answer", [(None, 1290), ('none', 1290),
                                                    ('default', 1233),
                                                    ('hard', 1101), ('hardest', 1101),
                                                    (1, 1290), (100, 1278), (2096639, 1101)])
def test_bitmasking(quality_bitmask, answer):
    """Test whether the bitmasking behaves like it should"""
    tpf = KeplerTargetPixelFile(filename_tpf_one_center, quality_bitmask=quality_bitmask)
    with warnings.catch_warnings():
        # Ignore "LightCurve contains NaN times" warnings triggered by liberal masks
        warnings.simplefilter("ignore", LightkurveWarning)
        lc = tpf.to_lightcurve()
    assert len(lc.flux) == answer


def test_wcs():
    """Test the wcs property."""
    for tpf in [KeplerTargetPixelFile(filename_tpf_one_center),
                TessTargetPixelFile(filename_tess)]:
        w = tpf.wcs
        ra, dec = tpf.get_coordinates()
        assert ra.shape == tpf.shape
        assert dec.shape == tpf.shape
        assert type(w).__name__ == 'WCS'


@pytest.mark.remote_data
@pytest.mark.parametrize("method", [("moments"), ("quadratic")])
def test_wcs_tabby(method):
    '''Test the centroids from Tabby's star against simbad values'''
    tpf = KeplerTargetPixelFile(TABBY_TPF)
    tpf.wcs
    ra, dec = tpf.get_coordinates(0)
    col, row = tpf.estimate_centroids(method=method)
    col = col.value - tpf.column
    row = row.value - tpf.row
    y, x = int(np.round(col[0])), int(np.round(row[1]))
    # Compare with RA and Dec from Simbad
    assert np.isclose(ra[x, y], 301.5643971, 1e-4)
    assert np.isclose(dec[x, y], 44.4568869, 1e-4)


def test_centroid_methods_consistency():
    """Are the centroid methods consistent for a well behaved target?"""
    pixels = read(filename_synthetic_flat)
    centr_moments = pixels.estimate_centroids(method='moments')
    centr_quadratic = pixels.estimate_centroids(method='quadratic')
    # check that the maximum relative difference doesnt exceed 1%
    assert np.max(np.abs(centr_moments[0] - centr_quadratic[0]) / centr_moments[0]) < 1e-2
    assert np.max(np.abs(centr_moments[1] - centr_quadratic[1]) / centr_moments[1]) < 1e-2


def test_properties():
    """Test the short-hand properties."""
    tpf = KeplerTargetPixelFile(filename_tpf_all_zeros)
    assert(tpf.channel == tpf.hdu[0].header['CHANNEL'])
    assert(tpf.module == tpf.hdu[0].header['MODULE'])
    assert(tpf.output == tpf.hdu[0].header['OUTPUT'])
    assert(tpf.ra == tpf.hdu[0].header['RA_OBJ'])
    assert(tpf.dec == tpf.hdu[0].header['DEC_OBJ'])
    assert_array_equal(tpf.flux.value, tpf.hdu[1].data['FLUX'][tpf.quality_mask])
    assert_array_equal(tpf.flux_err.value, tpf.hdu[1].data['FLUX_ERR'][tpf.quality_mask])
    assert_array_equal(tpf.flux_bkg.value, tpf.hdu[1].data['FLUX_BKG'][tpf.quality_mask])
    assert_array_equal(tpf.flux_bkg_err.value, tpf.hdu[1].data['FLUX_BKG_ERR'][tpf.quality_mask])
    assert_array_equal(tpf.quality, tpf.hdu[1].data['QUALITY'][tpf.quality_mask])
    assert(tpf.campaign == tpf.hdu[0].header['CAMPAIGN'])
    assert(tpf.quarter is None)


def test_repr():
    """Do __str__ and __repr__ work?"""
    for tpf in [KeplerTargetPixelFile(filename_tpf_all_zeros),
                TessTargetPixelFile(filename_tess)]:
        str(tpf)
        repr(tpf)


def test_to_lightcurve():
    for tpf in [KeplerTargetPixelFile(filename_tpf_all_zeros),
                TessTargetPixelFile(filename_tess)]:
        tpf.to_lightcurve()
        tpf.to_lightcurve(aperture_mask=None)
        tpf.to_lightcurve(aperture_mask='all')
        lc = tpf.to_lightcurve(aperture_mask='threshold')
        assert lc.time.scale == 'tdb'
        assert lc.label == tpf.hdu[0].header['OBJECT']
        if np.any(tpf.pipeline_mask):
            tpf.to_lightcurve(aperture_mask='pipeline')
        else:
            with pytest.raises(ValueError):
                tpf.to_lightcurve(aperture_mask='pipeline')


def test_bkg_lightcurve():
    for tpf in [KeplerTargetPixelFile(filename_tpf_all_zeros),
                TessTargetPixelFile(filename_tess)]:
        lc = tpf.get_bkg_lightcurve()
        lc = tpf.get_bkg_lightcurve(aperture_mask=None)
        lc = tpf.get_bkg_lightcurve(aperture_mask='all')
        assert lc.time.scale == 'tdb'
        assert lc.flux.shape == lc.flux_err.shape
        assert len(lc.time) == len(lc.flux)


def test_aperture_photometry():
    for tpf in [KeplerTargetPixelFile(filename_tpf_all_zeros),
                TessTargetPixelFile(filename_tess)]:
        tpf.extract_aperture_photometry()
        for mask in [None, 'all', 'default', 'threshold', 'background']:
            tpf.extract_aperture_photometry(aperture_mask=mask)
        if np.any(tpf.pipeline_mask):
            tpf.extract_aperture_photometry(aperture_mask='pipeline')
        else:
            with pytest.raises(ValueError):
                tpf.extract_aperture_photometry(aperture_mask='pipeline')


def test_tpf_to_fits():
    """Can we write a TPF back to a fits file?"""
    for tpf in [KeplerTargetPixelFile(filename_tpf_all_zeros),
                TessTargetPixelFile(filename_tess)]:
        # `delete=False` is necessary to enable writing to the file on Windows
        # but it means we have to clean up the tmp file ourselves
        tmp = tempfile.NamedTemporaryFile(delete=False)
        try:
            tpf.to_fits(tmp.name)
        finally:
            tmp.close()
            os.remove(tmp.name)


def test_tpf_factory():
    """Can we create TPFs using TargetPixelFileFactory?"""
    from lightkurve.targetpixelfile import FactoryError

    factory = TargetPixelFileFactory(n_cadences=10, n_rows=6, n_cols=8)
    flux_0 = np.ones((6, 8))
    factory.add_cadence(frameno=0, flux=flux_0,
                        header={'TSTART': 0, 'TSTOP': 10})
    flux_9 = 3 * np.ones((6, 8))
    factory.add_cadence(frameno=9, flux=flux_9,
                        header={'TSTART': 90, 'TSTOP': 100})

    # You shouldn't be able to build a TPF like this...because TPFs shouldn't
    # have extensions where time stamps are duplicated (here frames 1-8 will have)
    # time stamp zero
    with pytest.warns(LightkurveWarning, match='identical TIME values'):
        tpf = factory.get_tpf()
    [factory.add_cadence(frameno=i, flux=flux_0,
                         header={'TSTART': i*10, 'TSTOP': (i*10)+10})
     for i in np.arange(2, 9)]

    # This should fail because the time stamps of the images are not in order...
    with pytest.warns(LightkurveWarning, match='chronological order'):
        tpf = factory.get_tpf()

    [factory.add_cadence(frameno=i, flux=flux_0,
                         header={'TSTART': i*10, 'TSTOP': (i*10)+10})
     for i in np.arange(1, 9)]

    # This should pass
    tpf = factory.get_tpf(hdu0_keywords={'TELESCOP':'TESS'})

    assert_array_equal(tpf.flux[0].value, flux_0)
    assert_array_equal(tpf.flux[9].value, flux_9)

    tpf = factory.get_tpf(hdu0_keywords={'TELESCOP':'Kepler'})

    assert_array_equal(tpf.flux[0].value, flux_0)
    assert_array_equal(tpf.flux[9].value, flux_9)
    assert(tpf.time[0].value == 5)
    assert(tpf.time[9].value == 95)

    # Can you add the WRONG sized frame?
    flux_wrong = 3 * np.ones((6, 9))
    with pytest.raises(FactoryError):
        factory.add_cadence(frameno=2, flux=flux_wrong,
                            header={'TSTART': 90, 'TSTOP': 100})

    # Can you add the WRONG cadence?
    flux_wrong = 3 * np.ones((6, 8))
    with pytest.raises(FactoryError):
        factory.add_cadence(frameno=11, flux=flux_wrong,
                            header={'TSTART': 90, 'TSTOP': 100})

    # Can we add our own keywords?
    tpf = factory.get_tpf(hdu0_keywords={'creator': 'Christina TargetPixelFileWriter', 'TELESCOP':'TESS'})
    assert tpf.get_keyword('CREATOR') == 'Christina TargetPixelFileWriter'


def _create_image_array(header=None, shape=(5, 5)):
    """Helper function for tests below."""
    if header is None:
        header = fits.Header()
    images = []
    for i in range(5):
        header['TSTART'] = i
        header['TSTOP'] = i + 1
        images.append(fits.ImageHDU(data=np.ones(shape), header=header))
    return images

def test_tpf_from_images():
    """Basic tests of tpf.from_fits_images()"""
    # Not without a wcs...
    with pytest.raises(Exception):
        TargetPixelFile.from_fits_images(_create_image_array(), size=(3, 3),
                                               position=SkyCoord(-234.75, 8.3393, unit='deg'))

    # Make a fake WCS based on astropy.docs...
    w = wcs.WCS(naxis=2)
    w.wcs.crpix = [-234.75, 8.3393]
    w.wcs.cdelt = np.array([-0.066667, 0.066667])
    w.wcs.crval = [0, -90]
    w.wcs.ctype = ["RA---AIR", "DEC--AIR"]
    w.wcs.set_pv([(2, 1, 45.0)])
    pixcrd = np.array([[0, 0], [24, 38], [45, 98]], np.float_)
    header = w.to_header()
    header['CRVAL1P'] = 10
    header['CRVAL2P'] = 20
    ra, dec = 268.21686048, -73.66991904

    # Now this should work.
    images = _create_image_array(header=header)
    tpf = TargetPixelFile.from_fits_images(images, size=(3, 3),
                                                 position=SkyCoord(ra, dec, unit=(u.deg, u.deg)))
    assert isinstance(tpf, TargetPixelFile)


    with warnings.catch_warnings():
        # Some cards are too long -- to be investigated.
        warnings.simplefilter("ignore", VerifyWarning)
        # Can we write the output to disk?
        # `delete=False` is necessary below to enable writing to the file on Windows
        # but it means we have to clean up the tmp file ourselves
        tmp = tempfile.NamedTemporaryFile(delete=False)
        try:
            tpf.to_fits(tmp.name)
        finally:
            tmp.close()
            os.remove(tmp.name)

        # Can we read in a list of file names or a list of HDUlists?
        hdus = []
        tmpfile_names = []
        for im in images:
            tmpfile = tempfile.NamedTemporaryFile(delete=False)
            tmpfile_names.append(tmpfile.name)
            hdu = fits.HDUList([fits.PrimaryHDU(), im])
            hdu.writeto(tmpfile.name)
            hdus.append(hdu)

        # Should be able to run with a list of file names
        tpf_tmpfiles = TargetPixelFile.from_fits_images(tmpfile_names,
                            size=(3, 3),
                            position=SkyCoord(ra, dec, unit=(u.deg, u.deg)))

        # Should be able to run with a list of HDUlists
        tpf_hdus = TargetPixelFile.from_fits_images(hdus,
                            size=(3, 3),
                            position=SkyCoord(ra, dec, unit=(u.deg, u.deg)))

        # Clean up the temporary files we created
        for filename in tmpfile_names:
            try:
                os.remove(filename)
            except PermissionError:
                pass  # This appears to happen on Windows


def test_tpf_wcs_from_images():
    """Test to see if tpf.from_fits_images() output a tpf with WCS in the header"""
    # Not without a wcs...
    with pytest.raises(Exception):
        TargetPixelFile.from_fits_images(_create_image_array(), size=(3, 3),
                                               position=SkyCoord(-234.75, 8.3393, unit='deg'))

    # Make a fake WCS based on astropy.docs...
    w = wcs.WCS(naxis=2)
    w.wcs.crpix = [0., 0.]
    w.wcs.cdelt = np.array([0.001111, 0.001111])
    w.wcs.crval = [23.2334, 45.2333]
    w.wcs.ctype = ["RA---TAN", "DEC--TAN"]
    header = w.to_header()
    header['CRVAL1P'] = 10
    header['CRVAL2P'] = 20
    ra, dec = 23.2336, 45.235

    # Now this should work.
    tpf = TargetPixelFile.from_fits_images(_create_image_array(header=header), size=(3, 3),
                                                 position=SkyCoord(ra, dec, unit=(u.deg, u.deg)))
    assert tpf.hdu[1].header['1CRPX5'] != UNDEFINED
    assert tpf.hdu[1].header['1CTYP5'] == 'RA---TAN'
    assert tpf.hdu[1].header['2CTYP5'] == 'DEC--TAN'
    assert tpf.hdu[1].header['1CRPX5'] != UNDEFINED
    assert tpf.hdu[1].header['2CRPX5'] != UNDEFINED
    assert tpf.hdu[1].header['1CUNI5'] == 'deg'
    assert tpf.hdu[1].header['2CUNI5'] == 'deg'
    with warnings.catch_warnings():
        # Ignore the warning: "PC1_1 = a floating-point value was expected."
        warnings.simplefilter("ignore", AstropyWarning)
        assert tpf.wcs.to_header()['CDELT1'] == w.wcs.cdelt[0]


def test_properties2(capfd):
    '''Test if the describe function produces an output.
    The output is 1870 characters at the moment, but we might add more properties.'''
    tpf = KeplerTargetPixelFile(filename_tpf_all_zeros)
    tpf.show_properties()
    out, err = capfd.readouterr()
    assert len(out) > 1000


def test_interact():
    """Test the Jupyter notebook interact() widget."""
    for tpf in [KeplerTargetPixelFile(filename_tpf_one_center),
                TessTargetPixelFile(filename_tess)]:
        tpf.interact()


@pytest.mark.remote_data
def test_interact_sky():
    """Test the Jupyter notebook interact() widget."""
    for tpf in [KeplerTargetPixelFile(filename_tpf_one_center),
                TessTargetPixelFile(filename_tess)]:
        tpf.interact_sky()


def test_get_models():
    """Can we obtain PRF and TPF models?"""
    tpf = KeplerTargetPixelFile(filename_tpf_all_zeros, quality_bitmask=None)
    with warnings.catch_warnings():
        # Ignore "RuntimeWarning: All-NaN slice encountered"
        warnings.simplefilter("ignore", RuntimeWarning)
        tpf.get_model()
        tpf.get_prf_model()


@pytest.mark.remote_data
def test_tess_simulation():
    """Can we read simulated TESS data?"""
    tpf = TessTargetPixelFile(TESS_SIM)
    assert tpf.mission == 'TESS'
    assert tpf.time.scale == 'tdb'
    assert tpf.flux.shape == tpf.flux_err.shape
    tpf.wcs
    col, row = tpf.estimate_centroids()
    # Regression test for https://github.com/KeplerGO/lightkurve/pull/236
    assert (tpf.time.value == 0).sum() == 0


def test_threshold_aperture_mask():
    """Does the threshold mask work?"""
    tpf = KeplerTargetPixelFile(filename_tpf_one_center)
    tpf.plot(aperture_mask='threshold')
    lc = tpf.to_lightcurve(aperture_mask=tpf.create_threshold_mask(threshold=1))
    assert (lc.flux.value == 1).all()
    # The TESS file shows three pixel regions above a 2-sigma threshold;
    # let's make sure the `reference_pixel` argument allows them to be selected.
    tpf = TessTargetPixelFile(filename_tess)
    assert tpf.create_threshold_mask(threshold=2.).sum() == 25
    assert tpf.create_threshold_mask(threshold=2., reference_pixel='center').sum() == 25
    assert tpf.create_threshold_mask(threshold=2., reference_pixel=None).sum() == 28
    assert tpf.create_threshold_mask(threshold=2., reference_pixel=(5, 0)).sum() == 2
    # A mask which contains zero-flux pixels should work without crashing
    tpf = KeplerTargetPixelFile(filename_tpf_all_zeros)
    assert tpf.create_threshold_mask().sum() == 9


def test_tpf_tess():
    """Does a TESS Sector 1 TPF work?"""
    tpf = TessTargetPixelFile(filename_tess, quality_bitmask=None)
    assert tpf.mission == 'TESS'
    assert tpf.targetid == 25155310
    assert tpf.sector == 1
    assert tpf.camera == 4
    assert tpf.ccd == 1
    assert tpf.pipeline_mask.sum() == 9
    assert tpf.background_mask.sum() == 30
    lc = tpf.to_lightcurve()
    assert isinstance(lc, TessLightCurve)
    assert_array_equal(lc.time, tpf.time)
    assert tpf.time.scale == 'tdb'
    assert tpf.flux.shape == tpf.flux_err.shape
    tpf.wcs
    col, row = tpf.estimate_centroids()


@pytest.mark.parametrize("tpf_type", [KeplerTargetPixelFile, TessTargetPixelFile])
def test_tpf_slicing(tpf_type):
    """Test indexing and slicing of TargetPixelFile objects."""
    with warnings.catch_warnings():
        # Ignore the "TELESCOP is not equal to TESS" warning
        warnings.simplefilter("ignore", LightkurveWarning)
        tpf = tpf_type(filename_tpf_one_center)

    assert tpf[0].time == tpf.time[0]
    assert tpf[-1].time == tpf.time[-1]
    assert tpf[5:10].shape == tpf.flux[5:10].shape
    assert tpf[0].targetid == tpf.targetid
    assert_array_equal(tpf[tpf.time < tpf.time[5]].time, tpf.time[0:5])

    frame = tpf[5]
    assert frame.shape[0] == 1
    assert frame.shape[1:] == tpf.shape[1:]
    assert_array_equal(frame.time[0], tpf.time[5])
    assert_array_equal(frame.flux[0], tpf.flux[5])

    frames = tpf[100:200]
    assert frames.shape[0] == 100
    assert frames.shape[1:] == tpf.shape[1:]
    assert_array_equal(frames.time, tpf.time[100:200])
    assert_array_equal(frames.flux, tpf.flux[100:200])


def test_endianness():
    """Regression test for https://github.com/KeplerGO/lightkurve/issues/188"""
    tpf = KeplerTargetPixelFile(filename_tpf_one_center)
    tpf.to_lightcurve().to_pandas().describe()


def test_get_keyword():
    tpf = KeplerTargetPixelFile(filename_tpf_one_center)
    assert tpf.get_keyword("TELESCOP") == "Kepler"
    assert tpf.get_keyword("TTYPE1", hdu=1) == "TIME"
    assert tpf.get_keyword("DOESNOTEXIST", default=5) == 5


def test_cutout():
    """Test tpf.cutout() function."""
    for tpf in [KeplerTargetPixelFile(filename_tpf_one_center),
                TessTargetPixelFile(filename_tess, quality_bitmask=None)]:
        ntpf = tpf.cutout(size=2)
        assert ntpf.flux[0].shape == (2, 2)
        assert ntpf.flux_err[0].shape == (2, 2)
        assert ntpf.flux_bkg[0].shape == (2, 2)
        ntpf = tpf.cutout((0, 0), size=3)
        ntpf = tpf.cutout(size=(1, 2))
        assert ntpf.flux.shape[1] == 2
        assert ntpf.flux.shape[2] == 1
        ntpf = tpf.cutout(SkyCoord(tpf.ra, tpf.dec, unit='deg'), size=2)
        ntpf = tpf.cutout(size=2)
        assert np.product(ntpf.flux.shape[1:]) == 4
        assert ntpf.targetid == '{}_CUTOUT'.format(tpf.targetid)


def test_aperture_photometry_nan():
    """Regression test for #648.

    When FLUX or FLUX_ERR is entirely NaN in a TPF, the resulting light curve
    should report NaNs in that cadence rather than zero."""
    tpf = read(filename_tpf_one_center)
    tpf.hdu[1].data['FLUX'][2] = np.nan
    tpf.hdu[1].data['FLUX_ERR'][2] = np.nan
    lc = tpf.to_lightcurve(aperture_mask='all')
    assert ~np.isnan(lc.flux[1])
    assert ~np.isnan(lc.flux_err[1])
    assert np.isnan(lc.flux[2])
    assert np.isnan(lc.flux_err[2])


@pytest.mark.remote_data
def test_SSOs():
    # TESS test
    tpf = TessTargetPixelFile(asteroid_TPF)
    result = tpf.query_solar_system_objects() # default cadence_mask = 'outliers'
    assert(result is None) # the TPF has only data for 1 epoch. The lone time is removed as outlier
    result = tpf.query_solar_system_objects(cadence_mask='all', cache=False)
    assert(len(result) == 1)
    result = tpf.query_solar_system_objects(cadence_mask=np.asarray([True]), cache=False)
    assert(len(result) == 1)
    result = tpf.query_solar_system_objects(cadence_mask=[True], cache=False)
    assert(len(result) == 1)
    result = tpf.query_solar_system_objects(cadence_mask=(True), cache=False)
    assert(len(result) == 1)
    result, mask = tpf.query_solar_system_objects(cadence_mask=np.asarray([True]), cache=True, return_mask=True)
    assert(len(mask) == len(tpf.flux))
    try:
        result = tpf.query_solar_system_objects(cadence_mask='str-not-supported', cache=False)
        pytest.fail("Unsupported cadence_mask should have thrown Error")
    except ValueError:
        pass


def test_get_header():
    """Test the basic functionality of ``tpf.get_header()``"""
    tpf = read(filename_tpf_one_center)
    assert tpf.get_header()['CHANNEL'] == tpf.get_keyword("CHANNEL")
    assert tpf.get_header(0)['MISSION'] == tpf.get_keyword("MISSION")
    assert tpf.get_header(ext=2)['EXTNAME'] == "APERTURE"
    # ``tpf.header`` is deprecated
    with pytest.warns(LightkurveDeprecationWarning, match='deprecated'):
        tpf.header


def test_plot_pixels():
    tpf = KeplerTargetPixelFile(filename_tpf_one_center)
    tpf.plot_pixels()
    tpf.plot_pixels(normalize=True)
    tpf.plot_pixels(periodogram=True)
    tpf.plot_pixels(periodogram=True, nyquist_factor=0.5)
    tpf.plot_pixels(aperture_mask='all')
    tpf.plot_pixels(aperture_mask=tpf.pipeline_mask)
    tpf.plot_pixels(aperture_mask=tpf.create_threshold_mask())
    tpf.plot_pixels(show_flux=True)
    tpf.plot_pixels(corrector_func=lambda x:x)
    plt.close('all')


@pytest.mark.remote_data
def test_missing_pipeline_mask():
    """Regression test for #791.

    TPFs produced by TESSCut contain an empty pipeline mask.  When the pipeline
    mask is missing or empty, we want `to_lightcurve()` to fall back on the
    'threshold' mask by default, to avoid creating a light curve based on zero pixels."""
    tpf = search_tesscut("Proxima Cen", sector=12).download(cutout_size=1)
    lc = tpf.to_lightcurve()
    assert np.isfinite(lc.flux).any()
    assert lc.meta.get('APERTURE_MASK', None) == 'threshold'

    with pytest.raises(ValueError):
        # if aperture_mask is explicitly set as pipeline,
        # the logic will throw an error as it is missing in the TPF
        lc = tpf.to_lightcurve(aperture_mask='pipeline')


def test_cutout_quality_masking():
    """Regression test for #813: Does tpf.cutout() maintain the quality mask?"""
    tpf = read(filename_tpf_one_center, quality_bitmask=8192)
    tpfcut = tpf.cutout()
    assert(len(tpf) == len(tpfcut))


def test_parse_numeric_aperture_masks():
    """Regression test for #694: float or int aperture masks should be
    interpreted as boolean masks."""
    tpf = read(filename_tpf_one_center)
    mask = tpf._parse_aperture_mask(np.zeros(tpf.shape[1:], dtype=float))
    assert(mask.dtype == bool)
    mask = tpf._parse_aperture_mask(np.zeros(tpf.shape[1:], dtype=int))
    assert(mask.dtype == bool)


def test_tpf_meta():
    """Can we access meta data using tpf.meta?"""
    tpf = read(filename_tpf_one_center)
    assert tpf.meta.get('MISSION') == 'K2'
    assert tpf.meta['MISSION'] == 'K2'
    assert tpf.meta.get('mission', None) is None  # key is case in-sensitive
    assert tpf.meta.get('CHANNEL') == 45
    # ensure meta is read-only view of the underlying self.hdu[0].header
    with pytest.raises(TypeError):
        tpf.meta['CHANNEL'] = 44
    with pytest.raises(TypeError):
        tpf.meta['KEY-NEW'] = 44


def test_estimate_background():
    """Verifies tpf.estimate_background()."""
    # Create a TPF with 100 electron/second in every pixel
    tpf = read(filename_tpf_all_zeros) + 100.
    # The resulting background should be 100 e/s/pixel
    bg = tpf.estimate_background(aperture_mask='all')
    assert_array_equal(bg.flux.value, 100)
    assert bg.flux.unit == tpf.flux.unit / u.pixel



"""Defines a `PLDCorrector` class which provides a simple way to correct a
light curve by utilizing the pixel time series data contained within the
target's own Target Pixel File.

`PLDCorrector` builds upon `RegressionCorrector` by correlating the light curve
against a design matrix composed of the following elements:
* A background light curve to capture the dominant scattered light systematics.
* Background-corrected pixel time series to capture any residual systematics.
* Splines to capture the target's intrinsic variability.
"""
import logging
import warnings
from itertools import combinations_with_replacement as multichoose

import numpy as np
import matplotlib.pyplot as plt

from astropy.utils.decorators import deprecated, deprecated_renamed_argument

from .designmatrix import DesignMatrix, DesignMatrixCollection, \
                          SparseDesignMatrixCollection
from .regressioncorrector import RegressionCorrector
from .designmatrix import create_spline_matrix, create_sparse_spline_matrix
from .. import MPLSTYLE
from ..utils import LightkurveDeprecationWarning


log = logging.getLogger(__name__)


__all__ = ['PLDCorrector', 'TessPLDCorrector']


class PLDCorrector(RegressionCorrector):
    r"""Implements the Pixel Level Decorrelation (PLD) systematics removal method.

    Special case of `.RegressionCorrector` where the `.DesignMatrix` is
    composed of background-corrected pixel time series.

    The design matrix also contains columns representing a spline in time
    design to capture the intrinsic, long-term variability of the target.

    Pixel Level Decorrelation (PLD) was developed by [1]_ to remove
    systematic noise caused by spacecraft jitter for the Spitzer
    Space Telescope. It was adapted to K2 data by [2]_ and [3]_
    for the EVEREST pipeline [4]_.
    For a detailed description and implementation of PLD, please refer to
    these references. Lightkurve provides a reference implementation
    of PLD that is less sophisticated than EVEREST, but is suitable
    for quick-look analyses and detrending experiments.
    Our simple implementation of PLD is performed by first calculating the
    noise model for each cadence in time. This function goes up to arbitrary
    order, and is represented by

    .. math::
        m_i = \sum_l a_l \frac{f_{il}}{\sum_k f_{ik}} + \sum_l \sum_m b_{lm} \frac{f_{il}f_{im}}{\left( \sum_k f_{ik} \right)^2} + ...

    where
      - :math:`m_i` is the noise model at time :math:`t_i`
      - :math:`f_{il}` is the flux in the :math:`l^\text{th}` pixel at time :math:`t_i`
      - :math:`a_l` is the first-order PLD coefficient on the linear term
      - :math:`b_{lm}` is the second-order PLD coefficient on the :math:`l^\text{th}`,
        :math:`m^\text{th}` pixel pair
    We perform Principal Component Analysis (PCA) to reduce the number of
    vectors in our final model to limit the set to best capture instrumental
    noise. With a PCA-reduced set of vectors, we can construct a design matrix
    containing fractional pixel fluxes.
    To solve for the PLD model, we need to minimize the difference squared

    .. math::
        \chi^2 = \sum_i \frac{(y_i - m_i)^2}{\sigma_i^2},

    where :math:`y_i` is the observed flux value at time :math:`t_i`, by solving

    .. math::
        \frac{\partial \chi^2}{\partial a_l} = 0.

    The design matrix also contains columns representing a spline in time
    design to capture the intrinsic, long-term variability of the target.

    Examples
    --------
    Download the pixel data for GJ 9827 and obtain a PLD-corrected light curve:

    >>> import lightkurve as lk
    >>> tpf = lk.search_targetpixelfile("GJ9827").download() # doctest: +SKIP
    >>> corrector = tpf.to_corrector('pld') # doctest: +SKIP
    >>> lc = corrector.correct() # doctest: +SKIP
    >>> lc.plot() # doctest: +SKIP

    However, the above example will over-fit the small transits!
    It is necessary to mask the transits using `corrector.correct(cadence_mask=...)`.

    References
    ----------
    .. [1] Deming et al. (2015), ads:2015ApJ...805..132D.
        (arXiv:1411.7404)
    .. [2] Luger et al. (2016), ads:2016AJ....152..100L
        (arXiv:1607.00524)
    .. [3] Luger et al. (2018), ads:2018AJ....156...99L
        (arXiv:1702.05488)
    .. [4] EVEREST pipeline webpage, https://rodluger.github.io/everest
    """
    def __init__(self, tpf, aperture_mask=None):
        if aperture_mask is None:
            aperture_mask = tpf.create_threshold_mask(3)
        self.aperture_mask = aperture_mask
        lc = tpf.to_lightcurve(aperture_mask=aperture_mask)
        # Remove cadences that have NaN flux (cf. #874). We don't simply call
        # `lc.remove_nans()` here because we need to mask both lc & tpf.
        nan_mask = np.isnan(lc.flux)
        lc = lc[~nan_mask]
        self.tpf = tpf[~nan_mask]
        super().__init__(lc=lc)

    def __repr__(self):
        return 'PLDCorrector (ID: {})'.format(self.lc.label)

    def create_design_matrix(self, pld_order=3, pca_components=16, pld_aperture_mask=None,
                             background_aperture_mask='background', spline_n_knots=None,
                             spline_degree=3, normalize_background_pixels=None,
                             sparse=False):
        """Returns a `.DesignMatrixCollection` containing a `DesignMatrix` object
        for the background regressors, the PLD pixel component regressors, and
        the spline regressors.

        If the parameters `pld_order` and `pca_components` are None, their
        value will be assigned based on the mission. K2 and TESS experience
        different dominant sources of noise, and require different defaults.
        For information about how the defaults were chosen, see Pull Request #746.

        Parameters
        ----------
        pld_order : int
            The order of Pixel Level De-correlation to be performed. First order
            (`n=1`) uses only the pixel fluxes to construct the design matrix.
            Higher order populates the design matrix with columns constructed
            from the products of pixel fluxes.
        pca_components : int or tuple of int
            Number of terms added to the design matrix for each order of PLD
            pixel fluxes. Increasing this value may provide higher precision
            at the expense of slower speed and/or overfitting.
            If performing PLD with `pld_order > 1`, `pca_components` can be
            a tuple containing the number of terms for each order of PLD.
            If a single int is passed, the same number of terms will be used
            for each order. If zero is passed, PCA will not be performed.
            Defaults to 16 for K2 and 8 for TESS.
        pld_aperture_mask : array-like, 'pipeline', 'all', 'threshold', or None
            A boolean array describing the aperture such that `True` means
            that the pixel will be used when selecting the PLD basis vectors.
            If `None` or `all` are passed in, all pixels will be used.
            If 'pipeline' is passed, the mask suggested by the official pipeline
            will be returned.
            If 'threshold' is passed, all pixels brighter than 3-sigma above
            the median flux will be used.
        spline_n_knots : int
            Number of knots in spline.
        spline_degree : int
            Polynomial degree of spline.
        sparse : bool
            Whether to create `SparseDesignMatrix`.

        Returns
        -------
        dm : `.DesignMatrixCollection`
            `.DesignMatrixCollection` containing pixel, background, and spline
            components.
        """
        # Validate the inputs
        pld_aperture_mask = self.tpf._parse_aperture_mask(pld_aperture_mask)
        self.pld_aperture_mask = pld_aperture_mask
        background_aperture_mask = self.tpf._parse_aperture_mask(background_aperture_mask)
        self.background_aperture_mask = background_aperture_mask

        if spline_n_knots is None:
            # Default to a spline per 50 data points
            spline_n_knots = int(len(self.lc) / 50)

        if sparse:
            DMC = SparseDesignMatrixCollection
            spline = create_sparse_spline_matrix
        else:
            DMC = DesignMatrixCollection
            spline = create_spline_matrix

        # We set the width of all coefficient priors to 10 times the standard
        # deviation to prevent the fit from going crazy.
        prior_sigma = np.nanstd(self.lc.flux.value) * 10

        # Flux normalize background components for K2 and not for TESS by default
        bkg_pixels = self.tpf.flux[:, background_aperture_mask].reshape(len(self.tpf.flux), -1)
        if normalize_background_pixels:
            bkg_flux = np.nansum(self.tpf.flux[:, background_aperture_mask], -1)
            bkg_pixels = np.array([r / f for r, f in zip(bkg_pixels, bkg_flux)])
        else:
            bkg_pixels = bkg_pixels.value

        # Remove NaNs
        bkg_pixels = np.array([r[np.isfinite(r)] for r in bkg_pixels])

        # Create background design matrix
        dm_bkg = DesignMatrix(bkg_pixels, name='background')
        # Apply PCA
        dm_bkg = dm_bkg.pca(pca_components)
        # Set prior sigma to 10 * standard deviation
        dm_bkg.prior_sigma = np.ones(dm_bkg.shape[1]) * prior_sigma

        # Create a design matric containing splines plus a constant
        dm_spline = spline(self.lc.time.value,
                           n_knots=spline_n_knots,
                           degree=spline_degree).append_constant()
        # Set prior sigma to 10 * standard deviation
        dm_spline.prior_sigma = np.ones(dm_spline.shape[1]) * prior_sigma

        # Create a PLD matrix if there are pixels in the pld_aperture_mask
        if np.sum(pld_aperture_mask) != 0:
            # Flux normalize the PLD components
            pld_pixels = self.tpf.flux[:, pld_aperture_mask].reshape(len(self.tpf.flux), -1)
            pld_pixels = np.array([r / f for r, f in zip(pld_pixels, self.lc.flux.value)])
            # Remove NaNs
            pld_pixels = np.array([r[np.isfinite(r)] for r in pld_pixels])

            # Use the DesignMatrix infrastructure to apply PCA to the regressors.
            regressors_dm = DesignMatrix(pld_pixels)
            if pca_components > 0:
                regressors_dm = regressors_dm.pca(pca_components)
            regressors_pld = regressors_dm.values

            # Create a DesignMatrix for each PLD order
            all_pld = []
            for order in range(1, pld_order+1):
                reg_n = np.product(list(multichoose(regressors_pld.T, order)), axis=1).T
                pld_n = DesignMatrix(reg_n,
                                     prior_sigma=np.ones(reg_n.shape[1]) * prior_sigma / reg_n.shape[1],
                                     name=f"pld_order_{order}")
                # Apply PCA.
                if pca_components > 0:
                    pld_n = pld_n.pca(pca_components)
                    # Calling pca() resets the priors, so we set them again.
                    pld_n.prior_sigma = np.ones(pld_n.shape[1]) * prior_sigma / pca_components
                all_pld.append(pld_n)

            # Create the collection of DesignMatrix objects.
            # DesignMatrix 1 contains the PLD pixel series
            dm_pixels = DesignMatrixCollection(all_pld).to_designmatrix(name='pixel_series')

            with warnings.catch_warnings():
                warnings.filterwarnings('ignore', message='.*Not all matrices are `SparseDesignMatrix` objects..*')
                dm_collection = DMC([dm_pixels, dm_bkg, dm_spline])
        else:
            with warnings.catch_warnings():
                warnings.filterwarnings('ignore', message='.*Not all matrices are `SparseDesignMatrix` objects..*')
                dm_collection = DMC([dm_bkg, dm_spline])
        return dm_collection

    @deprecated_renamed_argument('n_pca_terms', 'pca_components', '2.0', warning_type=LightkurveDeprecationWarning)
    @deprecated_renamed_argument('use_gp', None, '2.0', warning_type=LightkurveDeprecationWarning)
    @deprecated_renamed_argument('gp_timescale', None, '2.0', warning_type=LightkurveDeprecationWarning)
    @deprecated_renamed_argument('aperture_mask', None, '2.0', warning_type=LightkurveDeprecationWarning)
    def correct(self, pld_order=None, pca_components=None, pld_aperture_mask=None,
                background_aperture_mask='background', spline_n_knots=None,
                spline_degree=5, normalize_background_pixels=None, restore_trend=True,
                sparse=False, cadence_mask=None, sigma=5, niters=5, propagate_errors=False,
                use_gp=None, gp_timescale=None, aperture_mask=None):
        """Returns a systematics-corrected light curve.

        If the parameters `pld_order` and `pca_components` are None, their
        value will be assigned based on the mission. K2 and TESS experience
        different dominant sources of noise, and require different defaults.
        For information about how the defaults were chosen, see PR #746 at
        https://github.com/KeplerGO/lightkurve/pull/746#issuecomment-658458270

        Parameters
        ----------
        pld_order : int
            The order of Pixel Level De-correlation to be performed. First order
            (`n=1`) uses only the pixel fluxes to construct the design matrix.
            Higher order populates the design matrix with columns constructed
            from the products of pixel fluxes. Default 3 for K2 and 1 for TESS.
        pca_components : int
            Number of terms added to the design matrix for each order of PLD
            pixel fluxes. Increasing this value may provide higher precision
            at the expense of slower speed and/or overfitting.
        pld_aperture_mask : array-like, 'pipeline', 'all', 'threshold', or None
            A boolean array describing the aperture such that `True` means
            that the pixel will be used when selecting the PLD basis vectors.
            If `None` or `all` are passed in, all pixels will be used.
            If 'pipeline' is passed, the mask suggested by the official pipeline
            will be returned.
            If 'threshold' is passed, all pixels brighter than 3-sigma above
            the median flux will be used.
        spline_n_knots : int
            Number of knots in spline.
        spline_degree : int
            Polynomial degree of spline.
        restore_trend : bool
            Whether to restore the long term spline trend to the light curve.
        sparse : bool
            Whether to create `SparseDesignMatrix`.
        cadence_mask : np.ndarray of bools (optional)
            Mask, where True indicates a cadence that should be used.
        sigma : int (default 5)
            Standard deviation at which to remove outliers from fitting
        niters : int (default 5)
            Number of iterations to fit and remove outliers
        propagate_errors : bool (default False)
            Whether to propagate the uncertainties from the regression. Default is False.
            Setting to True will increase run time, but will sample from multivariate normal
            distribution of weights.
        use_gp, gp_timescale : DEPRECATED
            As of Lightkurve v2.0 PLDCorrector uses splines instead of Gaussian Processes.
        aperture_mask : DEPRECATED
            As of Lightkurve v2.0 the `aperture_mask` parameter needs to be
            passed to the class constructor.

        Returns
        -------
        clc : `.LightCurve`
            Noise-corrected `.LightCurve`.
        """
        self.restore_trend = restore_trend

        # Set mission-specific values for pld_order and pca_components
        if pld_order is None:
            if self.tpf.meta.get('MISSION') == 'K2':
                pld_order = 3
            else:
                pld_order = 1
        if pca_components is None:
            if self.tpf.meta.get('MISSION') == 'K2':
                pca_components = 16
            else:
                pca_components = 3
        if pld_aperture_mask is None:
            if self.tpf.meta.get('MISSION') == 'K2':
                # K2 noise is dominated by motion
                pld_aperture_mask = 'threshold'
            else:
                # TESS noise is dominated by background
                pld_aperture_mask = 'empty'
        if normalize_background_pixels is None:
            if self.tpf.meta.get('MISSION') == 'K2':
                normalize_background_pixels = True
            else:
                normalize_background_pixels = False

        dm = self.create_design_matrix(pld_aperture_mask=pld_aperture_mask,
                                       background_aperture_mask=background_aperture_mask,
                                       pld_order=pld_order,
                                       pca_components=pca_components,
                                       spline_n_knots=spline_n_knots,
                                       spline_degree=spline_degree,
                                       normalize_background_pixels=normalize_background_pixels,
                                       sparse=sparse)

        clc = super().correct(dm, cadence_mask=cadence_mask, sigma=sigma,
                              niters=niters, propagate_errors=propagate_errors)
        if restore_trend:
            clc += (self.diagnostic_lightcurves['spline']
                    - np.median(self.diagnostic_lightcurves['spline'].flux))
        return clc

    def diagnose(self):
        """Returns diagnostic plots to assess the most recent call to `correct()`.
        If `correct()` has not yet been called, a ``ValueError`` will be raised.

        Returns
        -------
        `~matplotlib.axes.Axes`
            The matplotlib axes object.
        """
        if not getattr(self, 'corrected_lc'):
            raise ValueError('You need to call the `correct()` method '
                             'before you can call `diagnose()`.')
        names = self.diagnostic_lightcurves.keys()

        # Plot the right version of corrected light curve
        if self.restore_trend:
            clc = self.corrected_lc \
                  + self.diagnostic_lightcurves['spline'] \
                  - np.median(self.diagnostic_lightcurves['spline'].flux)
        else:
            clc = self.corrected_lc

        uncorr_cdpp = self.lc.estimate_cdpp()
        corr_cdpp = clc.estimate_cdpp()

        # Get y-axis limits
        ylim = [min(min(self.lc.flux.value), min(clc.flux.value)),
                max(max(self.lc.flux.value), max(clc.flux.value))]

        # Use lightkurve plotting style
        with plt.style.context(MPLSTYLE):
            # Plot background model
            _, axs = plt.subplots(3, figsize=(10, 9), sharex=True)
            ax = axs[0]
            self.lc.plot(ax=ax, normalize=False, clip_outliers=True,
                         label=f'uncorrected ({uncorr_cdpp:.0f})')
            ax.set_xlabel('')
            ax.set_ylim(ylim)  # use same ylim for all plots

            # Plot pixel and spline components
            ax = axs[1]
            clc.plot(ax=ax, normalize=False, alpha=0.4,
                     label=f'corrected ({corr_cdpp:.0f})')
            for key, color in zip(names, ['dodgerblue', 'r', 'C1']):
                if key in ['background', 'spline', 'pixel_series']:
                    tmplc = self.diagnostic_lightcurves[key] \
                            - np.median(self.diagnostic_lightcurves[key].flux) \
                            + np.median(self.lc.flux)
                    tmplc.plot(ax=ax, c=color)
            ax.set_xlabel('')
            ax.set_ylim(ylim)

            # Plot final corrected light curve with outliers marked
            ax = axs[2]
            self.lc.plot(ax=ax, normalize=False, alpha=0.2,
                         label=f'uncorrected ({uncorr_cdpp:.0f})')
            clc[self.outlier_mask].scatter(normalize=False, c='r', marker='x',
                                           s=10, label='outlier_mask', ax=ax)
            clc[~self.cadence_mask].scatter(normalize=False, c='dodgerblue',
                                            marker='x', s=10, label='~cadence_mask',
                                            ax=ax)
            clc.plot(normalize=False, ax=ax, c='k',
                     label=f'corrected ({corr_cdpp:.0f})')
            ax.set_ylim(ylim)
        return axs

    def diagnose_masks(self):
        """Show different aperture masks used by PLD in the most recent call to
        `correct()`. If `correct()` has not yet been called, a ``ValueError``
        will be raised.

        Returns
        -------
        `~matplotlib.axes.Axes`
            The matplotlib axes object.
        """
        if not hasattr(self, 'corrected_lc'):
            raise ValueError('You need to call the `correct()` method '
                             'before you can call `diagnose()`.')

        # Use lightkurve plotting style
        with plt.style.context(MPLSTYLE):
            _, axs = plt.subplots(1, 3, figsize=(12, 4), sharey=True)
            # Show light curve aperture mask
            ax = axs[0]
            self.tpf.plot(ax=ax, show_colorbar=False,
                          aperture_mask=self.aperture_mask,
                          title='aperture_mask')
            # Show PLD pixel mask
            ax = axs[1]
            self.tpf.plot(ax=ax, show_colorbar=False,
                          aperture_mask=self.pld_aperture_mask,
                          title='pld_aperture_mask')
            ax = axs[2]
            self.tpf.plot(ax=ax, show_colorbar=False,
                          aperture_mask=self.background_aperture_mask,
                          title='background_aperture_mask')
        return axs


# `TessPLDCorrector` was briefly introduced in Lightkurve v1.9
# but was removed in v2.0 in favor of a single generic `PLDCorrector`.
@deprecated("2.0", alternative="PLDCorrector",
            warning_type=LightkurveDeprecationWarning)
class TessPLDCorrector(PLDCorrector):
    pass



"""Unit tests for the `RegressionCorrector` class."""
import warnings

import numpy as np
from numpy.testing import assert_almost_equal
import pandas as pd
import pytest

from ... import LightCurve, LightkurveWarning
from .. import RegressionCorrector, DesignMatrix


def test_regressioncorrector_priors():
    """This test will fit a design matrix containing the column vectors
    a=[1, 1] and b=[1, 2] to a light curve with flux=[5, 10].

    The best coefficients for this problem are [0, 5] because 0*a + 5*b == flux,
    however we will verify that changing the priors will yield different
    solutions.
    """
    lc1 = LightCurve(flux=[5, 10])
    lc2 = LightCurve(flux=[5, 10], flux_err=[1, 1])
    design_matrix = DesignMatrix(pd.DataFrame({'a':[1, 1], 'b':[1, 2]}))
    for dm in [design_matrix, design_matrix.to_sparse()]:
        for lc in [lc1, lc2]:
            rc = RegressionCorrector(lc)

            # No prior
            rc.correct(dm)
            assert_almost_equal(rc.coefficients, [0, 5])

            # Strict prior centered on correct solution
            dm.prior_mu = [0, 5]
            dm.prior_sigma = [1e-6, 1e-6]
            rc.correct(dm)
            assert_almost_equal(rc.coefficients, [0, 5])

            # Strict prior centered on incorrect solution
            dm.prior_mu = [99, 99]
            dm.prior_sigma = [1e-6, 1e-6]
            rc.correct(dm)
            assert_almost_equal(rc.coefficients, [99, 99])

            # Wide prior centered on incorrect solution
            dm.prior_mu = [9, 9]
            dm.prior_sigma = [1e6, 1e6]
            rc.correct(dm)
            assert_almost_equal(rc.coefficients, [0, 5])

def test_sinusoid_noise():
    """Can we remove simple sinusoid noise added to a flat light curve?"""
    size = 100
    time = np.linspace(1, 100, size)
    true_flux = np.ones(size)
    noise = np.sin(time/5)
    # True light curve is flat, i.e. flux=1 at all time steps
    true_lc = LightCurve(time=time, flux=true_flux, flux_err=0.1*np.ones(size))
    # Noisy light curve has a sinusoid single added
    noisy_lc = LightCurve(time=time, flux=true_flux+noise, flux_err=true_lc.flux_err)
    design_matrix = DesignMatrix({'noise': noise,
                                  'offset': np.ones(len(time))},
                                  name='noise_model')

    for dm in [design_matrix, design_matrix.to_sparse()]:
        # Can we recover the true light curve?
        rc = RegressionCorrector(noisy_lc)
        corrected_lc = rc.correct(dm)
        assert_almost_equal(corrected_lc.normalize().flux, true_lc.flux)

        # Can we produce the diagnostic plot?
        rc.diagnose()

        # Does it work when we set priors?
        dm.prior_mu = [0.1, 0.1]
        dm.prior_sigma = [1e6, 1e6]
        corrected_lc = RegressionCorrector(noisy_lc).correct(dm)
        assert_almost_equal(corrected_lc.normalize().flux, true_lc.flux)

        # Does it work when `flux_err` isn't available?
        noisy_lc = LightCurve(time=time, flux=true_flux+noise)
        corrected_lc = RegressionCorrector(noisy_lc).correct(dm)
        assert_almost_equal(corrected_lc.normalize().flux, true_lc.flux)


def test_nan_input():
    # The following light curves should all raise ValueErrors because of NaNs
    with warnings.catch_warnings():
        # Instantiating light curves with NaN times will yield a warning
        warnings.simplefilter("ignore", LightkurveWarning)
        lcs = [LightCurve(flux=[5, 10], flux_err=[np.nan, 1]),
               LightCurve(flux=[np.nan, 10], flux_err=[1, 1])]

    # Passing these to RegressionCorrector should raise a ValueError
    for lc in lcs:
        with pytest.raises(ValueError):
            RegressionCorrector(lc)

    # However, we should be flexible with letting `flux_err` be all-NaNs,
    # because it is common for errors to be missing.
    lc = LightCurve(flux=[5, 10], flux_err=[np.nan, np.nan])
    RegressionCorrector(lc)


def test_zero_fluxerr():
    """Regression test for #668.

    Flux uncertainties smaller than or equal to zero (`lc.flux_err <= 0`) will
    trigger an invalid or non-finite matrix.  We expect `RegressionCorrector`
    to detect this and yield a graceful `ValueError`."""
    lc = LightCurve(flux=[5, 10], flux_err=[1, 0])
    with pytest.raises(ValueError):
        RegressionCorrector(lc)
    lc = LightCurve(flux=[5, 10], flux_err=[1, -10])
    with pytest.raises(ValueError):
        RegressionCorrector(lc)



"""Adds the BKJD and BTJD time format for use by Astropy's `Time` object."""
from astropy.time.formats import TimeNumeric, day_frac


class TimeBKJD(TimeNumeric):
    """
    Barycentric Kepler Julian Date time format.
    This represents the number of days since January 1, 2009 12:00:00 UTC.
    BKJD is the format in which times are recorded in Kepler data products.
    See Section 2.3.2 in the Kepler Archive Manual for details. 
    """
    name = 'bkjd'
    BKJDREF = 2454833  # Barycentric Kepler Julian Date offset

    def set_jds(self, val1, val2):
        self._check_scale(self._scale)  # Validate scale.
        jd1, jd2 = day_frac(val1, val2)
        jd1 += self.BKJDREF
        self.jd1, self.jd2 = day_frac(jd1, jd2)

    def to_value(self, **kwargs):
        jd1 = self.jd1 - self.BKJDREF
        jd2 = self.jd2
        return super().to_value(jd1=jd1, jd2=jd2, **kwargs)

    value = property(to_value)


class TimeBTJD(TimeNumeric):
    """
    Barycentric TESS Julian Date time format.
    This represents the number of days since JD 2457000.0.
    BTJD is the format in which times are recorded in TESS data products.
    """
    name = 'btjd'
    BTJDREF = 2457000  # Barycentric TESS Julian Date offset

    def set_jds(self, val1, val2):
        self._check_scale(self._scale)  # Validate scale.
        jd1, jd2 = day_frac(val1, val2)
        jd1 += self.BTJDREF
        self.jd1, self.jd2 = day_frac(jd1, jd2)

    def to_value(self, **kwargs):
        jd1 = self.jd1 - self.BTJDREF
        jd2 = self.jd2
        return super().to_value(jd1=jd1, jd2=jd2, **kwargs)

    value = property(to_value)



"""Read a generic FITS table containing a light curve."""
import logging

from astropy.io import fits
from astropy.table import Table
from astropy.time import Time
import numpy as np

from ..utils import validate_method
from ..lightcurve import LightCurve


log = logging.getLogger(__name__)


def read_generic_lightcurve(filename, flux_column,
                            quality_column='quality',
                            cadenceno_column='cadenceno',
                            centroid_col_column='mom_centr1',
                            centroid_row_column='mom_centr2',
                            time_format=None,
                            ext=1):
    """Generic helper function to convert a Kepler ot TESS light curve file
    into a generic `LightCurve` object.
    """
    if isinstance(filename, fits.HDUList):
        hdulist = filename  # Allow HDUList to be passed
    else:
        hdulist = fits.open(filename)

    # Raise an exception if the requested extension is invalid
    if isinstance(ext, str):
        validate_method(ext, supported_methods=[hdu.name.lower() for hdu in hdulist])
    tab = Table.read(hdulist[ext], format="fits")

    # Make sure the meta data also includes header fields from extension #0
    tab.meta.update(hdulist[0].header)

    tab.meta = {k : v for k, v in tab.meta.items()}

    # Some KEPLER files used to have a T column instead of TIME.
    if "T" in tab.colnames:
        tab.rename_column("T", "TIME")

    for colname in tab.colnames:
        # Ensure units have the correct astropy format
        # Speed-up: comparing units by their string representation is 1000x
        # faster than performing full-blown unit comparison
        unitstr = str(tab[colname].unit)
        if unitstr == 'e-/s':
            tab[colname].unit = 'electron/s'
        elif unitstr == 'pixels':
            tab[colname].unit = 'pixel'

        # Rename columns to lowercase
        tab.rename_column(colname, colname.lower())

    # We *have* to remove rows with TIME=NaN because the Astropy Time
    # object does not support the presence of NaNs.
    # Fortunately, such rows are always bad data.
    nans = np.isnan(tab['time'].data)
    if np.any(nans):
        log.debug('Ignoring {} rows with NaN times'.format(np.sum(nans)))
    tab = tab[~nans]

    # Prepare a special time column
    if not time_format:
        if hdulist[ext].header.get('BJDREFI') == 2454833:
            time_format = 'bkjd'
        elif hdulist[ext].header.get('BJDREFI') == 2457000:
            time_format = 'btjd'
        else:
            raise ValueError(f"Input file has unclear time format: {filename}")
    time = Time(tab['time'].data,
                scale=hdulist[ext].header.get('TIMESYS', 'tdb').lower(),
                format=time_format)
    tab.remove_column('time')

    # For backwards compatibility with Lightkurve v1.x,
    # we make sure standard columns and attributes exist.
    if 'flux' not in tab.columns:
        tab.add_column(tab[flux_column], name="flux", index=0)
    if 'flux_err' not in tab.columns and f"{flux_column}_err" in tab.columns:
        tab.add_column(tab[f"{flux_column}_err"], name="flux_err", index=1)
    if 'quality' not in tab.columns and quality_column in tab.columns:
        tab.add_column(tab[quality_column], name="quality", index=2)
    if 'cadenceno' not in tab.columns and cadenceno_column in tab.columns:
        tab.add_column(tab[cadenceno_column], name="cadenceno", index=3)
    if 'centroid_col' not in tab.columns and centroid_col_column in tab.columns:
        tab.add_column(tab[centroid_col_column], name="centroid_col", index=4)
    if 'centroid_row' not in tab.columns and centroid_row_column in tab.columns:
        tab.add_column(tab[centroid_row_column], name="centroid_row", index=5)

    tab.meta['LABEL'] = hdulist[0].header.get('OBJECT')
    tab.meta['MISSION'] = hdulist[0].header.get('MISSION', hdulist[0].header.get('TELESCOP'))
    tab.meta['RA'] = hdulist[0].header.get('RA_OBJ')
    tab.meta['DEC'] = hdulist[0].header.get('DEC_OBJ')
    tab.meta['FILENAME'] = filename

    return LightCurve(time=time, data=tab)



from __future__ import division, print_function

import pytest

from numpy.testing import assert_almost_equal

from ..lightcurvefile import KeplerLightCurveFile
from ..search import search_targetpixelfile

from .test_lightcurve import TABBY_Q8

@pytest.mark.remote_data
def test_to_corrector():
    """Does the tpf.to_corrector('pld') convenience method work?"""
    from .. import KeplerTargetPixelFile
    from .test_targetpixelfile import TABBY_TPF
    tpf = KeplerTargetPixelFile(TABBY_TPF)
    lc = tpf.to_corrector("pld").correct()
    assert len(lc.flux) == len(tpf.time)



import warnings

import pytest
from astropy.utils.data import get_pkg_data_filename
import matplotlib.pyplot as plt
import numpy as np
from numpy.testing import assert_array_equal

from ..lightcurve import LightCurve, KeplerLightCurve, TessLightCurve
from ..search import search_lightcurve
from ..targetpixelfile import KeplerTargetPixelFile, TessTargetPixelFile
from ..collections import LightCurveCollection, TargetPixelFileCollection
from ..utils import LightkurveWarning

filename_tpf_all_zeros = get_pkg_data_filename("data/test-tpf-all-zeros.fits")
filename_tpf_one_center = get_pkg_data_filename("data/test-tpf-non-zero-center.fits")


def test_collection_init():
    lc = LightCurve(time=np.arange(1, 5), flux=np.arange(1, 5), flux_err=np.arange(1, 5))
    lc2 = LightCurve(time=np.arange(10, 15), flux=np.arange(10, 15), flux_err=np.arange(10, 15))
    lcc = LightCurveCollection([lc, lc2])
    assert(len(lcc) == 2)
    assert(lcc.data == [lc, lc2])
    str(lcc)  # Does repr work?
    lcc.plot()
    plt.close('all')


def test_collection_append():
    """Does Collection.append() work?"""
    lc = LightCurve(time=np.arange(1, 5), flux=np.arange(1, 5),
                    flux_err=np.arange(1, 5), targetid=500)
    lc2 = LightCurve(time=np.arange(10, 15), flux=np.arange(10, 15),
                     flux_err=np.arange(10, 15), targetid=100)
    lcc = LightCurveCollection([lc])
    lcc.append(lc2)
    assert(len(lcc) == 2)

def test_collection_stitch():
    """Does Collection.stitch() work?"""
    lc = LightCurve(time=np.arange(1, 5), flux=np.ones(4))
    lc2 = LightCurve(time=np.arange(5, 16), flux=np.ones(11))
    lcc = LightCurveCollection([lc, lc2])
    lc_stitched = lcc.stitch()
    assert(len(lc_stitched.flux) == 15)
    lc_stitched2 = lcc.stitch(corrector_func=lambda x: x*2)
    assert_array_equal(lc_stitched.flux*2, lc_stitched2.flux)

def test_collection_getitem():
    """Tests Collection.__getitem__"""
    lc = LightCurve(time=np.arange(1, 5), flux=np.arange(1, 5),
                    flux_err=np.arange(1, 5), targetid=50000)
    lc2 = LightCurve(time=np.arange(10, 15), flux=np.arange(10, 15),
                     flux_err=np.arange(10, 15), targetid=120334)
    lcc = LightCurveCollection([lc])
    lcc.append(lc2)
    assert((lcc[0] == lc).all())
    assert((lcc[1] == lc2).all())
    with pytest.raises(IndexError):
        lcc[50]

def test_collection_getitem_by_boolean_array():
    """Tests Collection.__getitem__ , case the argument is a mask, i.e, indexed by boolean array"""
    lc0 = LightCurve(time=np.arange(1, 5), flux=np.arange(1, 5),
                    flux_err=np.arange(1, 5), targetid=50000)
    lc1 = LightCurve(time=np.arange(10, 15), flux=np.arange(10, 15),
                     flux_err=np.arange(10, 15), targetid=120334)
    lc2 = LightCurve(time=np.arange(15, 20), flux=np.arange(15, 20),
                     flux_err=np.arange(15, 20), targetid=23456)
    lcc = LightCurveCollection([lc0, lc1, lc2])

    lcc_f = lcc[[True, False, True]]
    assert(lcc_f.data == [lc0, lc2])
    assert(type(lcc_f), LightCurveCollection)

    # boundary case: 1 element
    lcc_f = lcc[[False, True, False]]
    assert(lcc_f.data == [lc1])
    # boundary case: no element
    lcc_f = lcc[[False, False, False]]
    assert(lcc_f.data == [])
    # other array-like input: tuple
    lcc_f = lcc[(True, False, True)]
    assert(lcc_f.data == [lc0, lc2])
    # other array-like input: ndarray
    lcc_f = lcc[np.array([True, False, True])]
    assert(lcc_f.data == [lc0, lc2])

    # boundary case: mask length not matching - shorter
    with pytest.raises(IndexError):
        lcc[[True, False]]

    # boundary case: mask length not matching - longer
    with pytest.raises(IndexError):
        lcc[[True, False, True, True]]

def test_collection_getitem_by_other_array():
    """Tests Collection.__getitem__ , case the argument an non-boolean array"""
    lc0 = LightCurve(time=np.arange(1, 5), flux=np.arange(1, 5),
                    flux_err=np.arange(1, 5), targetid=50000)
    lc1 = LightCurve(time=np.arange(10, 15), flux=np.arange(10, 15),
                     flux_err=np.arange(10, 15), targetid=120334)
    lc2 = LightCurve(time=np.arange(15, 20), flux=np.arange(15, 20),
                     flux_err=np.arange(15, 20), targetid=23456)
    lcc = LightCurveCollection([lc0, lc1, lc2])

    # case: an int array-like, follow ndarray behavior
    lcc_f = lcc[[2, 0]]
    assert(lcc_f.data == [lc2, lc0])
    lcc_f = lcc[np.array([2, 0])]
    assert(lcc_f.data == [lc2, lc0])
    # support other int types in np too
    lcc_f = lcc[np.array([np.int64(2), np.uint8(0)])]
    assert(lcc_f.data == [lc2, lc0])
    # boundary condition: True / False is interpreted as 1/0 in an bool/int mixed array-like
    lcc_f = lcc[[True, False, 2]]
    assert(lcc_f.data == [lc1, lc0, lc2])
    # boundary condition: some index is out of bound
    with pytest.raises(IndexError):
        lcc[[2, 99]]
    # boundary conditions: array-like of neither bool or int, follow ndarray behavior
    with pytest.raises(IndexError):
        lcc[['abc', 'def']]
    with pytest.raises(IndexError):
        lcc[[True, 'def']]

def test_collection_setitem():
    """Tests Collection. __setitem__"""
    lc = LightCurve(time=np.arange(1, 5), flux=np.arange(1, 5),
                    flux_err=np.arange(1, 5), targetid=50000)
    lc2 = LightCurve(time=np.arange(10, 15), flux=np.arange(10, 15),
                     flux_err=np.arange(10, 15), targetid=120334)
    lcc = LightCurveCollection([lc])
    lcc.append(lc2)
    lc3 = LightCurve(time=[1], targetid=55)
    lcc[1] = lc3
    assert(lcc[1].time == lc3.time)
    lcc.append(lc2)
    assert((lcc[2].time == lc2.time).all())
    with pytest.raises(IndexError):
        lcc[51] = 10


def test_tpfcollection():
    tpf = KeplerTargetPixelFile(filename_tpf_all_zeros)
    tpf2 = KeplerTargetPixelFile(filename_tpf_one_center)
    tpfc = TargetPixelFileCollection([tpf, tpf2])
    assert(len(tpfc) == 2)
    assert(tpfc.data == [tpf, tpf2])
    tpfc.append(tpf2)
    assert(len(tpfc) == 3)
    assert(tpfc[0] == tpf)
    assert(tpfc[1] == tpf2)
    assert(tpfc[2] == tpf2)
    with pytest.raises(IndexError):
        tpfc[51]
    # ensure index by boolean array also works for TPFs
    tpfc_f = tpfc[[False, True, True]]
    assert(tpfc_f.data == [tpf2, tpf2])
    assert(type(tpfc_f), TargetPixelFileCollection)
    # Test __setitem__
    tpf3 = KeplerTargetPixelFile(filename_tpf_one_center, targetid=55)
    tpfc[1] = tpf3
    assert(tpfc[1] == tpf3)
    tpfc.append(tpf2)
    assert(tpfc[2] == tpf2)
    str(tpfc)  # Regression test for #564


def test_tpfcollection_plot():
    tpf = KeplerTargetPixelFile(filename_tpf_all_zeros)
    tpf2 = KeplerTargetPixelFile(filename_tpf_one_center)
    # Does plotting work with 3 TPFs?
    coll = TargetPixelFileCollection([tpf, tpf2, tpf2])
    coll.plot()
    # Does plotting work with one TPF?
    coll = TargetPixelFileCollection([tpf])
    coll.plot()
    plt.close('all')


@pytest.mark.remote_data
def test_stitch_repr():
    """Regression test for #884."""
    lc = search_lightcurve("Pi Men", mission='TESS', sector=1).download()
    # The line below used to raise `ValueError: Unable to parse format string
    # "{:10d}" for entry "70445.0" in column "cadenceno"`
    LightCurveCollection((lc,lc)).stitch().__repr__()


def test_accessor_tess_sector():
    lc0 = TessLightCurve(time=np.arange(1, 5), flux=np.arange(1, 5),
                         flux_err=np.arange(1, 5), targetid=50000)
    lc0.meta['SECTOR'] = 14
    lc1 = TessLightCurve(time=np.arange(10, 15), flux=np.arange(10, 15),
                         flux_err=np.arange(10, 15), targetid=120334)
    lc1.meta['SECTOR'] = 26
    lcc = LightCurveCollection([lc0, lc1])
    assert((lcc.sector == [14, 26]).all())
    # The sector accessor can be used to generate boolean array
    # to support filter collection by sector
    assert(((lcc.sector == 26) == [False, True]).all())
    assert(((lcc.sector < 20) == [True, False]).all())

    # boundary condition: some lightcurve objects do not have sector
    lc2 = LightCurve(time=np.arange(15, 20), flux=np.arange(15, 20),
                     flux_err=np.arange(15, 20), targetid=23456)
    lcc.append(lc2)
    # expecting [14, 26, np.nan], need 2 asserts to do it.
    assert((lcc.sector[:-1] == [14, 26]).all())
    assert(np.isnan(lcc.sector[-1]))
    # The sector accessor can be used to generate boolean array
    # to support filter collection by sector
    assert(((lcc.sector == 26) == [False, True, False]).all())
    assert(((lcc.sector < 20) == [True, False, False]).all())

    # ensure it works for TPFs too.
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", LightkurveWarning)
        # Ignore "A Kepler data product is being opened using the `TessTargetPixelFile` class"
        # the test only cares about the SECTOR header that it sets.
        tpf = TessTargetPixelFile(filename_tpf_all_zeros)
        tpf.hdu[0].header['SECTOR'] = 23
        tpf2 = TessTargetPixelFile(filename_tpf_one_center)
        # tpf2 has no sector defined
        tpf3 = TessTargetPixelFile(filename_tpf_one_center)
        tpf3.hdu[0].header['SECTOR'] = 1
    tpfc = TargetPixelFileCollection([tpf, tpf2, tpf3])
    assert((tpfc.sector == [23, None, 1]).all())


def test_accessor_kepler_quarter():
    # scaled down version of tess sector test, as they share the same codepath
    lc0 = KeplerLightCurve(time=np.arange(1, 5), flux=np.arange(1, 5),
                           flux_err=np.arange(1, 5), targetid=50000)
    lc0.meta['QUARTER'] = 2
    lc1 = KeplerLightCurve(time=np.arange(10, 15), flux=np.arange(10, 15),
                           flux_err=np.arange(10, 15), targetid=120334)
    lc1.meta['QUARTER'] = 1
    lcc = LightCurveCollection([lc0, lc1])
    assert((lcc.quarter == [2, 1]).all())

    # ensure it works for TPFs too.
    tpf0 = KeplerTargetPixelFile(filename_tpf_all_zeros)
    tpf0.hdu[0].header['QUARTER'] = 2
    tpf1 = KeplerTargetPixelFile(filename_tpf_one_center)
    tpf1.hdu[0].header['QUARTER'] = 1
    tpfc = TargetPixelFileCollection([tpf0, tpf1])
    assert((tpfc.quarter == [2, 1]).all())


def test_accessor_k2_campaign():
    # scaled down version of tess sector test, as they share the same codepath
    lc0 = KeplerLightCurve(time=np.arange(1, 5), flux=np.arange(1, 5),
                           flux_err=np.arange(1, 5), targetid=50000)
    lc0.meta['CAMPAIGN'] = 2
    lc1 = KeplerLightCurve(time=np.arange(10, 15), flux=np.arange(10, 15),
                           flux_err=np.arange(10, 15), targetid=120334)
    lc1.meta['CAMPAIGN'] = 1
    lcc = LightCurveCollection([lc0, lc1])
    assert((lcc.campaign == [2, 1]).all())

    # ensure it works for TPFs too.
    tpf0 = KeplerTargetPixelFile(filename_tpf_all_zeros)
    tpf0.hdu[0].header['CAMPAIGN'] = 2
    tpf1 = KeplerTargetPixelFile(filename_tpf_one_center)
    tpf1.hdu[0].header['CAMPAIGN'] = 1
    tpfc = TargetPixelFileCollection([tpf0, tpf1])
    assert((tpfc.campaign == [2, 1]).all())



"""This module provides various helper functions."""
import logging
import sys
import os
import warnings
from functools import wraps

import matplotlib.pyplot as plt
from matplotlib.colors import LogNorm
import numpy as np
import pandas as pd
from tqdm import tqdm

import astropy
from astropy.utils.data import download_file
from astropy.units.quantity import Quantity
import astropy.units as u
from astropy.visualization import (PercentileInterval, ImageNormalize,
                                   SqrtStretch, LinearStretch)
from astropy.time import Time


log = logging.getLogger(__name__)


__all__ = ['LightkurveError', 'LightkurveWarning',
           'KeplerQualityFlags', 'TessQualityFlags',
           'bkjd_to_astropy_time', 'btjd_to_astropy_time',
           'show_citation_instructions']


class QualityFlags(object):
    """Abstract class"""

    STRINGS = {}
    OPTIONS = {}

    @classmethod
    def decode(cls, quality):
        """Converts a QUALITY value into a list of human-readable strings.

        This function takes the QUALITY bitstring that can be found for each
        cadence in Kepler/K2/TESS' pixel and light curve files and converts into
        a list of human-readable strings explaining the flags raised (if any).

        Parameters
        ----------
        quality : int
            Value from the 'QUALITY' column of a Kepler/K2/TESS pixel or lightcurve file.

        Returns
        -------
        flags : list of str
            List of human-readable strings giving a short description of the
            quality flags raised.  Returns an empty list if no flags raised.
        """
        # If passed an astropy quantity object, get the value
        if isinstance(quality, Quantity):
            quality = quality.value
        result = []
        for flag in cls.STRINGS.keys():
            if quality & flag > 0:
                result.append(cls.STRINGS[flag])
        return result

    @classmethod
    def create_quality_mask(cls, quality_array, bitmask=None):
        """Returns a boolean array which flags good cadences given a bitmask.

        This method is used by the readers of :class:`KeplerTargetPixelFile`
        and :class:`KeplerLightCurve` to initialize their `quality_mask`
        class attribute which is used to ignore bad-quality data.

        Parameters
        ----------
        quality_array : array of int
            'QUALITY' column of a Kepler target pixel or lightcurve file.
        bitmask : int or str
            Bitmask (int) or one of 'none', 'default', 'hard', or 'hardest'.

        Returns
        -------
        boolean_mask : array of bool
            Boolean array in which `True` means the data is of good quality.
        """
        # Return an array filled with `True` by default (i.e. ignore nothing)
        if bitmask is None:
            return np.ones(len(quality_array), dtype=bool)
        if isinstance(quality_array, u.Quantity):
            quality_array = quality_array.value
        # A few pre-defined bitmasks can be specified as strings
        if isinstance(bitmask, str):
            try:
                bitmask = cls.OPTIONS[bitmask]
            except KeyError:
                valid_options = tuple(cls.OPTIONS.keys())
                raise ValueError("quality_bitmask='{}' is not supported, "
                                 "expected one of {}"
                                 "".format(bitmask, valid_options))
        # The bitmask is applied using the bitwise AND operator
        quality_mask = (quality_array & bitmask) == 0
        # Log the quality masking as info or warning
        n_cadences = len(quality_array)
        n_cadences_masked = (~quality_mask).sum()
        percent_masked = 100. * n_cadences_masked / n_cadences
        logmsg = "{:.0f}% ({}/{}) of the cadences will be ignored due to the " \
                 "quality mask (quality_bitmask={})." \
                 "".format(percent_masked, n_cadences_masked, n_cadences, bitmask)
        if percent_masked > 20:
            log.warning("Warning: " + logmsg)
        else:
            log.info(logmsg)
        return quality_mask


class KeplerQualityFlags(QualityFlags):
    """
    This class encodes the meaning of the various Kepler QUALITY bitmask flags,
    as documented in the Kepler Archive Manual (Ref. [1], Table 2.3).

    References
    ----------
    .. [1] Kepler: A Search for Terrestrial Planets. Kepler Archive Manual.
        http://archive.stsci.edu/kepler/manuals/archive_manual.pdf
    """
    AttitudeTweak = 1
    SafeMode = 2
    CoarsePoint = 4
    EarthPoint = 8
    ZeroCrossing = 16
    Desat = 32
    Argabrightening = 64
    ApertureCosmic = 128
    ManualExclude = 256
    # Bit 2**10 = 512 is unused by Kepler
    SensitivityDropout = 1024
    ImpulsiveOutlier = 2048
    ArgabrighteningOnCCD = 4096
    CollateralCosmic = 8192
    DetectorAnomaly = 16384
    NoFinePoint = 32768
    NoData = 65536
    RollingBandInAperture = 131072
    RollingBandInMask = 262144
    PossibleThrusterFiring = 524288
    ThrusterFiring = 1048576

    #: DEFAULT bitmask identifies all cadences which are definitely useless.
    DEFAULT_BITMASK = (AttitudeTweak | SafeMode | CoarsePoint | EarthPoint |
                       Desat | ManualExclude | DetectorAnomaly | NoData | ThrusterFiring)
    #: HARD bitmask is conservative and may identify cadences which are useful.
    HARD_BITMASK = (DEFAULT_BITMASK | SensitivityDropout | ApertureCosmic |
                    CollateralCosmic | PossibleThrusterFiring)
    #: HARDEST bitmask identifies cadences with any flag set. Its use is not recommended.
    HARDEST_BITMASK = 2096639

    #: Dictionary which provides friendly names for the various bitmasks.
    OPTIONS = {'none': 0,
               'default': DEFAULT_BITMASK,
               'hard': HARD_BITMASK,
               'hardest': HARDEST_BITMASK}

    #: Pretty string descriptions for each flag
    STRINGS = {
        1: "Attitude tweak",
        2: "Safe mode",
        4: "Coarse point",
        8: "Earth point",
        16: "Zero crossing",
        32: "Desaturation event",
        64: "Argabrightening",
        128: "Cosmic ray in optimal aperture",
        256: "Manual exclude",
        1024: "Sudden sensitivity dropout",
        2048: "Impulsive outlier",
        4096: "Argabrightening on CCD",
        8192: "Cosmic ray in collateral data",
        16384: "Detector anomaly",
        32768: "No fine point",
        65536: "No data",
        131072: "Rolling band in optimal aperture",
        262144: "Rolling band in full mask",
        524288: "Possible thruster firing",
        1048576: "Thruster firing"
    }


class TessQualityFlags(QualityFlags):
    """
    This class encodes the meaning of the various TESS QUALITY bitmask flags,
    as documented in the TESS Data Products Description Document (Ref. [1], Table 26).

    References
    ----------
    .. [1] TESS Science Data Products Description Document (EXP-TESS-ARC-ICD-0014)
        https://archive.stsci.edu/missions/tess/doc/EXP-TESS-ARC-ICD-TM-0014.pdf
    """
    AttitudeTweak = 1
    SafeMode = 2
    CoarsePoint = 4
    EarthPoint = 8
    Argabrightening = 16
    Desat = 32
    ApertureCosmic = 64
    ManualExclude = 128
    Discontinuity = 256
    ImpulsiveOutlier = 512
    CollateralCosmic = 1024
    #: The first stray light flag is set manually by MIT based on visual inspection.
    Straylight = 2048
    #: The second stray light flag is set automatically by Ames/SPOC based on background level thresholds.
    Straylight2 = 4096

    #: DEFAULT bitmask identifies all cadences which are definitely useless.
    DEFAULT_BITMASK = (AttitudeTweak | SafeMode | CoarsePoint | EarthPoint |
                       Desat | ManualExclude)
    #: HARD bitmask is conservative and may identify cadences which are useful.
    HARD_BITMASK = (DEFAULT_BITMASK | ApertureCosmic |
                    CollateralCosmic | Straylight | Straylight2)
    #: HARDEST bitmask identifies cadences with any flag set. Its use is not recommended.
    HARDEST_BITMASK = 8191

    #: Dictionary which provides friendly names for the various bitmasks.
    OPTIONS = {'none': 0,
               'default': DEFAULT_BITMASK,
               'hard': HARD_BITMASK,
               'hardest': HARDEST_BITMASK}

    #: Pretty string descriptions for each flag
    STRINGS = {
        1: "Attitude tweak",
        2: "Safe mode",
        4: "Coarse point",
        8: "Earth point",
        16: "Argabrightening",
        32: "Desaturation event",
        64: "Cosmic ray in optimal aperture",
        128: "Manual exclude",
        256: "Discontinuity corrected",
        512: "Impulsive outlier",
        1024: "Cosmic ray in collateral data",
        2048: "Straylight",
        4096: "Straylight2"
    }

def channel_to_module_output(channel):
    """Returns a (module, output) pair given a CCD channel number.

    Parameters
    ----------
    channel : int
        Channel number

    Returns
    -------
    module, output : tuple of ints
        Module and Output number
    """
    if channel < 1 or channel > 88:
        raise ValueError("Channel number must be in the range 1-88.")
    lookup = _get_channel_lookup_array()
    lookup[:, 0] = 0
    modout = np.where(lookup == channel)
    return (modout[0][0], modout[1][0])


def module_output_to_channel(module, output):
    """Returns the CCD channel number for a given module and output pair.

    Parameters
    ----------
    module : int
        Module number
    output : int
        Output number

    Returns
    -------
    channel : int
        Channel number
    """
    if module < 1 or module > 26:
        raise ValueError("Module number must be in range 1-26.")
    if output < 1 or output > 4:
        raise ValueError("Output number must be 1, 2, 3, or 4.")
    return _get_channel_lookup_array()[module, output]


def _get_channel_lookup_array():
    """Returns a lookup table which maps (module, output) onto channel."""
    # In the array below, channel == array[module][output]
    # Note: modules 1, 5, 21, 25 are the FGS guide star CCDs.
    return np.array([
        [0,     0,    0,    0,    0],
        [1,    85,    0,    0,    0],
        [2,     1,    2,    3,    4],
        [3,     5,    6,    7,    8],
        [4,     9,   10,   11,   12],
        [5,    86,    0,    0,    0],
        [6,    13,   14,   15,   16],
        [7,    17,   18,   19,   20],
        [8,    21,   22,   23,   24],
        [9,    25,   26,   27,   28],
        [10,   29,   30,   31,   32],
        [11,   33,   34,   35,   36],
        [12,   37,   38,   39,   40],
        [13,   41,   42,   43,   44],
        [14,   45,   46,   47,   48],
        [15,   49,   50,   51,   52],
        [16,   53,   54,   55,   56],
        [17,   57,   58,   59,   60],
        [18,   61,   62,   63,   64],
        [19,   65,   66,   67,   68],
        [20,   69,   70,   71,   72],
        [21,   87,    0,    0,    0],
        [22,   73,   74,   75,   76],
        [23,   77,   78,   79,   80],
        [24,   81,   82,   83,   84],
        [25,   88,    0,    0,    0],
    ])


def running_mean(data, window_size):
    """Returns the moving average of an array `data`.

    Parameters
    ----------
    data : array of numbers
        The running mean will be computed on this data.
    window_size : int
        Window length used to compute the running mean.
    """
    if window_size > len(data):
        window_size = len(data)
    cumsum = np.cumsum(np.insert(data, 0, 0))
    return (cumsum[window_size:] - cumsum[:-window_size]) / float(window_size)


def bkjd_to_astropy_time(bkjd) -> Time:
    """Converts Kepler Barycentric Julian Day (BKJD) time values to an
    `astropy.time.Time` object.

    Kepler Barycentric Julian Day (BKJD) is a Julian day minus 2454833.0
    (UTC=January 1, 2009 12:00:00) and corrected to the arrival times
    at the barycenter of the Solar System.
    BKJD is the format in which times are recorded in the Kepler data products.
    The time is in the Barycentric Dynamical Time frame (TDB), which is a
    time system that is not affected by leap seconds.
    See Section 2.3.2 in the Kepler Archive Manual for details.

    Parameters
    ----------
    bkjd : float or array of floats
        Barycentric Kepler Julian Day.

    Returns
    -------
    time : `astropy.time.Time` object
        Resulting time object.
    """
    bkjd = np.atleast_1d(bkjd)
    # Some data products have missing time values;
    # we need to set these to zero or `Time` cannot be instantiated.
    bkjd[~np.isfinite(bkjd)] = 0
    return Time(bkjd, format='bkjd', scale='tdb')


def btjd_to_astropy_time(btjd) -> Time:
    """Converts TESS Barycentric Julian Day (BTJD) values to an
    `astropy.time.Time` object.

    TESS Barycentric Julian Day (BTJD) is a Julian day minus 2457000.0
    and corrected to the arrival times at the barycenter of the Solar System.
    BTJD is the format in which times are recorded in the TESS data products.
    The time is in the Barycentric Dynamical Time frame (TDB), which is a
    time system that is not affected by leap seconds.

    Parameters
    ----------
    btjd : float or array of floats
        Barycentric TESS Julian Day

    Returns
    -------
    time : `astropy.time.Time` object
        Resulting time object.
    """
    btjd = np.atleast_1d(btjd)
    btjd[~np.isfinite(btjd)] = 0
    return Time(btjd, format='btjd', scale='tdb')


def plot_image(image, ax=None, scale='linear', origin='lower',
               xlabel='Pixel Column Number', ylabel='Pixel Row Number',
               clabel='Flux ($e^{-}s^{-1}$)', title=None, show_colorbar=True,
               vmin=None, vmax=None, **kwargs):
    """Utility function to plot a 2D image

    Parameters
    ----------
    image : 2d array
        Image data.
    ax : `~matplotlib.axes.Axes`
        A matplotlib axes object to plot into. If no axes is provided,
        a new one will be generated.
    scale : str
        Scale used to stretch the colormap.
        Options: 'linear', 'sqrt', or 'log'.
    origin : str
        The origin of the coordinate system.
    xlabel : str
        Label for the x-axis.
    ylabel : str
        Label for the y-axis.
    clabel : str
        Label for the color bar.
    title : str or None
        Title for the plot.
    show_colorbar : bool
        Whether or not to show the colorbar
    vmin : float
        Minimum colorbar value. By default, the 2.5%-percentile is used.
    vmax : float
        Maximum colorbar value. By default, the 97.5%-percentile is used.
    kwargs : dict
        Keyword arguments to be passed to `matplotlib.pyplot.imshow`.

    Returns
    -------
    ax : `~matplotlib.axes.Axes`
        The matplotlib axes object.
    """
    if isinstance(image, u.Quantity):
        image = image.value
    if ax is None:
        _, ax = plt.subplots()

    if vmin is None or vmax is None:
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", RuntimeWarning)  # ignore image NaN values
            mask = np.nan_to_num(image) > 0
            if mask.any() > 0:
                vmin_default, vmax_default = PercentileInterval(95.).get_limits(image[mask])
            else:
                vmin_default, vmax_default = 0, 0
            if vmin is None:
                vmin = vmin_default
            if vmax is None:
                vmax = vmax_default

    norm = None
    if scale is not None:
        if scale == 'linear':
            norm = ImageNormalize(vmin=vmin, vmax=vmax, stretch=LinearStretch(), clip=False)
        elif scale == 'sqrt':
            norm = ImageNormalize(vmin=vmin, vmax=vmax, stretch=SqrtStretch(), clip=False)
        elif scale == 'log':
            # To use log scale we need to guarantee that vmin > 0, so that
            # we avoid division by zero and/or negative values.
            norm = LogNorm(vmin=max(vmin, sys.float_info.epsilon), vmax=vmax,
                           clip=True)
        else:
            raise ValueError("scale {} is not available.".format(scale))
    cax = ax.imshow(image, origin=origin, norm=norm, **kwargs)
    ax.set_xlabel(xlabel)
    ax.set_ylabel(ylabel)
    ax.set_title(title)
    if show_colorbar:
        cbar = plt.colorbar(cax, ax=ax, label=clabel)
        cbar.ax.yaxis.set_tick_params(tick1On=False, tick2On=False)
        cbar.ax.minorticks_off()
    return ax


class LightkurveError(Exception):
    """Class for Lightkurve exceptions."""
    pass


class LightkurveWarning(Warning):
    """Class for all Lightkurve warnings."""
    pass


class LightkurveDeprecationWarning(LightkurveWarning):
    """Class for all Lightkurve deprecation warnings."""
    pass


def suppress_stdout(f, *args, **kwargs):
    """A simple decorator to suppress function print outputs."""
    @wraps(f)
    def wrapper(*args, **kwargs):
        # redirect output to `null`
        with open(os.devnull, 'w') as devnull:
            old_out = sys.stdout
            sys.stdout = devnull
            try:
                return f(*args, **kwargs)
            # restore to default
            finally:
                sys.stdout = old_out
    return wrapper


def validate_method(method, supported_methods):
    """Raises a `ValueError` if a method is not supported.

    Parameters
    ----------
    method : str
        The method specified by the user.
    supported_methods : list of str
        The methods supported.  All method names must be lowercase.

    Returns
    -------
    method : str
        Will return the method name if it is supported.
    """
    method = method.lower()
    if method in supported_methods:
        return method
    raise ValueError("method '{}' is not supported; "
                     "must be one of {}".format(method, supported_methods))


def centroid_quadratic(data, mask=None):
    """Computes the quadratic estimate of the centroid in a 2d-array.

    This method will fit a simple 2D second-order polynomial
    $P(x, y) = a + bx + cy + dx^2 + exy + fy^2$
    to the 3x3 patch of pixels centered on the brightest pixel within
    the image.  This function approximates the core of the Point
    Spread Function (PSF) using a bivariate quadratic function, and returns
    the maximum (x, y) coordinate of the function using linear algebra.

    For the motivation and the details around this technique, please refer
    to Vakili, M., & Hogg, D. W. 2016, ArXiv, 1610.05873.

    Caveat: if the brightest pixel falls on the edge of the data array, the fit
    will tend to fail or be inaccurate.

    Parameters
    ----------
    data : 2D array
        The 2D input array representing the pixel values of the image.
    mask : array_like (bool), optional
        A boolean mask, with the same shape as `data`, where a **True** value
        indicates the corresponding element of data is masked.

    Returns
    -------
    column, row : tuple
        The coordinates of the centroid in column and row.  If the fit failed,
        then (NaN, NaN) will be returned.
    """
    if isinstance(data, u.Quantity):
        data = data.value
    # Step 1: identify the patch of 3x3 pixels (z_)
    # that is centered on the brightest pixel (xx, yy)
    if mask is not None:
        data = data * mask
    arg_data_max = np.nanargmax(data)
    yy, xx = np.unravel_index(arg_data_max, data.shape)
    # Make sure the 3x3 patch does not leave the TPF bounds
    if yy < 1:
        yy = 1
    if xx < 1:
        xx = 1
    if yy > (data.shape[0] - 2):
        yy = data.shape[0] - 2
    if xx > (data.shape[1] - 2):
        xx = data.shape[1] - 2

    z_ = data[yy-1:yy+2, xx-1:xx+2]

    # Next, we will fit the coefficients of the bivariate quadratic with the
    # help of a design matrix (A) as defined by Eqn 20 in Vakili & Hogg
    # (arxiv:1610.05873). The design matrix contains a
    # column of ones followed by pixel coordinates: x, y, x**2, xy, y**2.
    A = np.array([[1, -1, -1, 1,  1, 1],
                  [1,  0, -1, 0,  0, 1],
                  [1,  1, -1, 1, -1, 1],
                  [1, -1,  0, 1,  0, 0],
                  [1,  0,  0, 0,  0, 0],
                  [1,  1,  0, 1,  0, 0],
                  [1, -1,  1, 1, -1, 1],
                  [1,  0,  1, 0,  0, 1],
                  [1,  1,  1, 1,  1, 1]])
    # We also pre-compute $(A^t A)^-1 A^t$, cf. Eqn 21 in Vakili & Hogg.
    At = A.transpose()
    # In Python 3 this can become `Aprime = np.linalg.inv(At @ A) @ At`
    Aprime = np.matmul(np.linalg.inv(np.matmul(At, A)), At)

    # Step 2: fit the polynomial $P = a + bx + cy + dx^2 + exy + fy^2$
    # following Equation 21 in Vakili & Hogg.
    # In Python 3 this can become `Aprime @ z_.flatten()`
    a, b, c, d, e, f = np.matmul(Aprime, z_.flatten())

    # Step 3: analytically find the function maximum,
    # following https://en.wikipedia.org/wiki/Quadratic_function
    det = 4 * d * f - e ** 2
    if abs(det) < 1e-6:
        return np.nan, np.nan  # No solution
    xm = - (2 * f * b - c * e) / det
    ym = - (2 * d * c - b * e) / det
    return xx + xm, yy + ym


def _query_solar_system_objects(ra, dec, times, radius=0.1, location='kepler',
                                cache=True):
    """Returns a list of asteroids/comets given a position and time.

    This function relies on The Virtual Observatory Sky Body Tracker (SkyBot)
    service which can be found at http://vo.imcce.fr/webservices/skybot/

    Parameters
    ----------
    ra : float
        Right Ascension in degrees.
    dec : float
        Declination in degrees.
    times : array of float
        Times in Julian Date.
    radius : float
        Search radius in degrees.
    location : str
        Spacecraft location. Options include `'kepler'` and `'tess'`.
    cache : bool
        Whether to cache the search result. Default is True.

    Returns
    -------
    result : `pandas.DataFrame`
        DataFrame containing the list of known solar system objects at the
        requested time and location.
    """
    if (location.lower() == 'kepler') or (location.lower() == 'k2'):
        location = 'C55'
    elif location.lower() == 'tess':
        location = 'C57'

    url = 'http://vo.imcce.fr/webservices/skybot/skybotconesearch_query.php?'
    url += '-mime=text&'
    url += '-ra={}&'.format(ra)
    url += '-dec={}&'.format(dec)
    url += '-bd={}&'.format(radius)
    url += '-loc={}&'.format(location)

    df = None
    times = np.atleast_1d(times)
    for time in tqdm(times, desc='Querying for SSOs'):
        url_queried = url + 'EPOCH={}'.format(time)
        response = download_file(url_queried, cache=cache)
        if open(response).read(10) == '# Flag: -1':  # error code detected?
            raise IOError("SkyBot Solar System query failed.\n"
                          "URL used:\n" + url_queried + "\n"
                          "Response received:\n" + open(response).read())
        res = pd.read_csv(response, delimiter='|', skiprows=2)
        if len(res) > 0:
            res['epoch'] = time
            res.rename({'# Num ':'Num', ' Name ':'Name', ' Class ':'Class', ' Mv ':'Mv'}, inplace=True, axis='columns')
            res = res[['Num', 'Name', 'Class', 'Mv', 'epoch']].reset_index(drop=True)
            if df is None:
                df = res
            else:
                df = df.append(res)
    if df is not None:
        df.reset_index(drop=True)
    return df


def show_citation_instructions():
    """Show citation instructions."""
    from . import PACKAGEDIR, __citation__
    if not is_notebook():
        print(__citation__)
    else:
        from pathlib import Path
        from IPython.display import HTML
        import astroquery
        templatefile = Path(PACKAGEDIR, 'data', 'show_citation_instructions.html')
        template = open(templatefile, 'r').read()
        template = template.replace("LIGHTKURVE_CITATION", __citation__)
        template = template.replace("ASTROPY_CITATION", astropy.__citation__)
        template = template.replace("ASTROQUERY_CITATION", astroquery.__citation__)
        return HTML(template)


def _get_notebook_environment():
    """Returns 'jupyter', 'colab', or 'terminal'.

    One can detect whether or not a piece of Python is running by executing
    `get_ipython().__class__`, which returns the following result:

        * Jupyter notebook: `ipykernel.zmqshell.ZMQInteractiveShell`
        * Google colab: `google.colab._shell.Shell`
        * IPython terminal: `IPython.terminal.interactiveshell.TerminalInteractiveShell`
        * Python terminal: `NameError: name 'get_ipython' is not defined`
    """
    try:
        ipy = str(type(get_ipython())).lower()
        if 'zmqshell' in ipy:
            return 'jupyter'
        if 'colab' in ipy:
            return 'colab'
    except NameError:
        pass  # get_ipython() is not a builtin
    return 'terminal'


def is_notebook():
    """Returns `True` if we are running in a notebook."""
    if _get_notebook_environment() in ['jupyter', 'colab']:
        return True
    return False



""" cbvcorrector.py module unit tests
"""

import pytest
from numpy.testing import (assert_almost_equal, assert_array_equal,
                           assert_allclose, assert_raises)

import warnings
import numpy as np
import matplotlib
import matplotlib.pyplot as plt
import astropy.units as u
import pandas as pd
from astropy.table import Table
from astropy.time import Time

from ... import TessLightCurve, KeplerLightCurve
from ... import search_lightcurve
from ... import LightkurveWarning
from ..designmatrix import DesignMatrix
from ..cbvcorrector import download_kepler_cbvs, download_tess_cbvs, \
    CotrendingBasisVectors, KeplerCotrendingBasisVectors, \
    TessCotrendingBasisVectors


def test_CotrendingBasisVectors_nonretrieval():
    """ Tests CotrendingBasisVectors class without requiring remote data
    """

    #***
    # Constructor
    # Create some generic CotrendingBasisVectors objects

    # Generic CotrendingBasisVectors object
    dataTbl = Table([[1, 2, 3], [False, True, False], [2.0, 3.0, 4.0], [3.0, 4.0, 5.0]], 
        names=('CADENCENO', 'GAP', 'VECTOR_1', 'VECTOR_3'))
    cbvTime = Time([443.51090033, 443.53133457, 443.55176891], format='bkjd')
    cbvs = CotrendingBasisVectors(data=dataTbl, time=cbvTime)
    assert isinstance(cbvs, CotrendingBasisVectors)
    assert cbvs.cbv_indices == [1, 3]
    assert np.all(cbvs.time.value == [443.51090033, 443.53133457, 443.55176891])
    
    # Auto-initiate 'GAP' and 'CADENCENO'
    dataTbl = Table([[2.0, 3.0, 4.0], [3.0, 4.0, 5.0]], 
            names=('VECTOR_3', 'VECTOR_12'))
    cbvTime = Time([443.51090033, 443.53133457, 443.55176891], format='bkjd')
    cbvs = CotrendingBasisVectors(data=dataTbl, time=cbvTime)
    assert isinstance(cbvs, CotrendingBasisVectors)
    assert cbvs.cbv_indices == [3, 12]
    assert np.all(cbvs.gap_indicators == [False, False, False])
    assert np.all(cbvs.cadenceno == [0, 1, 2])

    
    #***
    # _to_designmatrix
    # Make sure CBVs are the columns in the returned 2-dim array
    dataTbl = Table([[1, 2, 3], [False, True, False], 
            [1.0, 2.0, 3.0], [4.0, 5.0, 6.0], [7.0, 8.0, 9.0]], 
            names=('CADENCENO', 'GAP', 'VECTOR_1', 'VECTOR_2', 'VECTOR_3'))
    cbvTime = Time([1569.44053967, 1569.44192856, 1569.44331746], format='btjd')
    cbvs = CotrendingBasisVectors(dataTbl, cbvTime)
    cbv_dm_name = 'test cbv set'
    # CBV index 5 does not exists and should be ingored
    cbv_designmatrix = cbvs.to_designmatrix(cbv_indices=[1,3,5], name=cbv_dm_name)
    assert cbv_designmatrix.shape == (3,2)
    assert np.all(cbv_designmatrix['VECTOR_1'] == np.array([1.0, 2.0, 3.0]))
    assert np.all(cbv_designmatrix['VECTOR_3'] == np.array([7.0, 8.0, 9.0]))
    assert cbv_designmatrix.name == cbv_dm_name
    # CBV #2 was not requested, so make sure it is not present
    with pytest.raises(KeyError):
        cbv_designmatrix['VECTOR_2']
    
    #***
    # plot
    ax = cbvs.plot(cbv_indices=[1,2], ax=None)
    assert isinstance(ax, matplotlib.axes.Axes)
    
    # There is no CBV # 5 so the third cbv_indices entry will be ignored
    ax = cbvs.plot(cbv_indices=[1,2,5], ax=ax)
    assert isinstance(ax, matplotlib.axes.Axes)
    
    # CBVs use 1-based indexing. Throw error if requesting CBV index 0
    with pytest.raises(ValueError):
        ax = cbvs.plot(cbv_indices=[0,1,2], ax=ax)

    # Only 'all' or specific CBV indices can be requested
    with pytest.raises(ValueError):
        ax = cbvs.plot('Doh!')

    
    #***
    # align
    # Set up some cadenceno such that both CBV is trimmed and NaNs inserted
    sample_lc = TessLightCurve(time=[1,2,3,4,6,7], flux=[1,2,3,4,6,7],
            flux_err=[0.1,0.1, 0.1, 0.1, 0.1, 0.1], cadenceno=[1,2,3,4,6,7])
    dataTbl = Table([[1, 2, 3, 5, 6], [False, True, False, False, False], 
            [1.0, 2.0, 3.0, 5.0, 6.0 ]], 
        names=('CADENCENO', 'GAP', 'VECTOR_1'))
    cbvTime = Time([1569.43915078, 1569.44053967, 1569.44192856,
           1569.44470635, 1569.44609524], format='btjd')
    cbvs = CotrendingBasisVectors(dataTbl, cbvTime)
    cbvs = cbvs.align(sample_lc)
    assert np.all(sample_lc.cadenceno == cbvs.cadenceno)
    assert len(cbvs.cadenceno) == 6
    assert len(sample_lc.flux) == 6
    assert np.all(cbvs.gap_indicators.value[[1,3,5]])
    # Ignore the warning in to_designmatric due to a low rank matrix
    with warnings.catch_warnings():
        # Instantiating light curves with NaN times will yield a warning
        warnings.simplefilter("ignore", LightkurveWarning)
        cbv_designmatrix = cbvs.to_designmatrix(cbv_indices=[1])
    assert np.all(cbv_designmatrix['VECTOR_1'][[0,1,2,4]] == [1.0, 2.0, 3.0, 6.0])
    assert np.all(np.isnan(cbv_designmatrix['VECTOR_1'][[3,5]]))

    #***
    # interpolate
    nLcCadences = 20
    xLc = np.linspace(0.0, 2*np.pi, num=nLcCadences)
    sample_lc = TessLightCurve(time=xLc, flux=np.sin(xLc), flux_err=np.full(nLcCadences, 0.1),
            cadenceno=np.arange(nLcCadences))
    nCbvCadences = 10
    xCbv = np.linspace(0.0, 2*np.pi, num=nCbvCadences)
    dataTbl = Table([np.arange(nCbvCadences), np.full(nCbvCadences, False), 
            np.cos(xCbv), np.sin(xCbv+np.pi*0.125)], 
            names=('CADENCENO', 'GAP', 'VECTOR_1', 'VECTOR_2'))
    cbvTime = Time(xCbv, format='btjd')
    cbvs = CotrendingBasisVectors(dataTbl, cbvTime)
    cbv_interpolated = cbvs.interpolate(sample_lc, extrapolate=False)
    assert np.all(cbv_interpolated.time.value == sample_lc.time.value)
    # Extrapolation test
    xCbv = np.linspace(0.0, 1.5*np.pi, num=nCbvCadences)
    dataTbl = Table([np.arange(nCbvCadences), np.full(nCbvCadences, False), 
            np.cos(xCbv), np.sin(xCbv+np.pi*0.125)], 
            names=('CADENCENO', 'GAP', 'VECTOR_1', 'VECTOR_2'))
    cbvTime = Time(xCbv, format='btjd')
    cbvs = CotrendingBasisVectors(dataTbl, cbvTime)
    cbv_interpolated = cbvs.interpolate(sample_lc, extrapolate=False)
    assert np.all(np.isnan(
        cbv_interpolated['VECTOR_1'].value[
            np.nonzero(cbv_interpolated.time.value > 1.5*np.pi)[0]]))
    cbv_interpolated = cbvs.interpolate(sample_lc, extrapolate=True)
    assert np.all(np.logical_not(np.isnan(
        cbv_interpolated['VECTOR_1'].value[
            np.nonzero(cbv_interpolated.time.value > 1.5*np.pi)[0]])))
    
@pytest.mark.remote_data
def test_cbv_retrieval():
    """ Tests reading in some CBVs from MAST

        This indirectly tests the classes KeplerCotrendingBasisVectors and
        TessCotrendingBasisVectors

    """

    cbvs = download_tess_cbvs(sector=10, camera=2, ccd=4, cbv_type = 'SingleScale')
    assert isinstance(cbvs, TessCotrendingBasisVectors)
    ax = cbvs.plot([1,2,4,6,8])
    assert isinstance(ax, matplotlib.axes.Axes)
    assert cbvs.mission == 'TESS'
    assert cbvs.cbv_type == 'SingleScale'
    assert cbvs.band is None
    assert cbvs.sector == 10
    assert cbvs.camera == 2
    assert cbvs.ccd == 4
    
    cbvs = download_tess_cbvs(sector=10, camera=2, ccd=4, cbv_type = 'MultiScale', band=2)
    assert isinstance(cbvs, TessCotrendingBasisVectors)
    ax = cbvs.plot('all')
    assert isinstance(ax, matplotlib.axes.Axes)
    assert cbvs.band == 2
    
    cbvs = download_tess_cbvs(sector=8, camera=3, ccd=1, cbv_type = 'Spike')
    assert isinstance(cbvs, TessCotrendingBasisVectors)
    ax = cbvs.plot('all')
    assert isinstance(ax, matplotlib.axes.Axes)
    
    # No band specified for MultiScale, this should error
    with pytest.raises(AssertionError):
        cbvs = download_tess_cbvs(sector=10, camera=2, ccd=4, cbv_type = 'MultiScale')
    # Band specified for SingleScale, this should also error
    with pytest.raises(AssertionError):
        cbvs = download_tess_cbvs(sector=10, camera=2, ccd=4, cbv_type = 'SingleScale', band=2)
    # Improper CBV type request
    with pytest.raises(Exception):
        cbvs = download_tess_cbvs(sector=10, camera=2, ccd=4, cbv_type = 'SuperSingleScale')

    cbvs = download_kepler_cbvs(mission='Kepler', quarter=8, module=16, output=4)
    assert isinstance(cbvs, KeplerCotrendingBasisVectors)
    ax = cbvs.plot('all')
    assert isinstance(ax, matplotlib.axes.Axes)
    assert cbvs.mission == 'Kepler'
    assert cbvs.cbv_type == 'SingleScale'
    assert cbvs.quarter == 8
    assert cbvs.campaign is None
    assert cbvs.module == 16
    assert cbvs.output == 4

    cbvs = download_kepler_cbvs(mission='K2', campaign=15, channel=24)
    assert isinstance(cbvs, KeplerCotrendingBasisVectors)
    ax = cbvs.plot('all')
    assert isinstance(ax, matplotlib.axes.Axes)
    assert cbvs.mission == 'K2'
    assert cbvs.cbv_type == 'SingleScale'
    assert cbvs.quarter is None
    assert cbvs.campaign == 15
    assert cbvs.module == 8
    assert cbvs.output == 4



import pytest
import warnings

import numpy as np
from numpy.testing import assert_almost_equal, assert_array_equal

from ..utils import KeplerQualityFlags, TessQualityFlags
from ..utils import module_output_to_channel, channel_to_module_output
from ..utils import LightkurveWarning
from ..utils import running_mean, validate_method
from ..utils import bkjd_to_astropy_time, btjd_to_astropy_time
from ..utils import centroid_quadratic
from ..utils import show_citation_instructions
from ..lightcurve import LightCurve


def test_channel_to_module_output():
    assert channel_to_module_output(1) == (2, 1)
    assert channel_to_module_output(42) == (13, 2)
    assert channel_to_module_output(84) == (24, 4)
    assert channel_to_module_output(33) == (11, 1)
    with pytest.raises(ValueError):
        channel_to_module_output(0)  # Invalid channel


def test_module_output_to_channel():
    assert module_output_to_channel(2, 1) == 1
    assert module_output_to_channel(13, 2) == 42
    assert module_output_to_channel(24, 4) == 84
    assert module_output_to_channel(11, 1) == 33
    with pytest.raises(ValueError):
        module_output_to_channel(0, 1)  # Invalid module
    with pytest.raises(ValueError):
        module_output_to_channel(2, 0)  # Invalid output


def test_running_mean():
    assert_almost_equal(running_mean([1, 2, 3], window_size=1), [1, 2, 3])
    assert_almost_equal(running_mean([1, 2, 3], window_size=2), [1.5, 2.5])
    assert_almost_equal(running_mean([2, 2, 2], window_size=3), [2])
    assert_almost_equal(running_mean([3, 4, 5], window_size=20), [4])


def test_quality_flag_decoding_kepler():
    """Can the QUALITY flags be parsed correctly?"""
    flags = list(KeplerQualityFlags.STRINGS.items())
    for key, value in flags:
        assert KeplerQualityFlags.decode(key)[0] == value
    # Can we recover combinations of flags?
    assert KeplerQualityFlags.decode(flags[5][0] + flags[7][0]) == [flags[5][1], flags[7][1]]
    assert KeplerQualityFlags.decode(flags[3][0] + flags[4][0] + flags[5][0]) \
        == [flags[3][1], flags[4][1], flags[5][1]]


def test_quality_flag_decoding_tess():
    """Can the QUALITY flags be parsed correctly?"""
    flags = list(TessQualityFlags.STRINGS.items())
    for key, value in flags:
        assert TessQualityFlags.decode(key)[0] == value
    # Can we recover combinations of flags?
    assert TessQualityFlags.decode(flags[5][0] + flags[7][0]) == [flags[5][1], flags[7][1]]
    assert TessQualityFlags.decode(flags[3][0] + flags[4][0] + flags[5][0]) \
        == [flags[3][1], flags[4][1], flags[5][1]]


def test_quality_flag_decoding_quantity_object():
    """Can a QUALITY flag that is a astropy quantity object be parsed correctly?

    This is a regression test for https://github.com/KeplerGO/lightkurve/issues/804
    """
    from astropy.units.quantity import Quantity
    flags = list(TessQualityFlags.STRINGS.items())
    for key, value in flags:
        assert TessQualityFlags.decode(Quantity(key, dtype='int32'))[0] == value
    # Can we recover combinations of flags?
    assert TessQualityFlags.decode(Quantity(flags[5][0], dtype='int32') + \
        Quantity(flags[7][0], dtype='int32')) == [flags[5][1], flags[7][1]]
    assert TessQualityFlags.decode(Quantity(flags[3][0], dtype='int32') + \
        Quantity(flags[4][0], dtype='int32') + Quantity(flags[5][0], dtype='int32')) \
        == [flags[3][1], flags[4][1], flags[5][1]]


def test_quality_mask():
    """Can we create a quality mask using KeplerQualityFlags?"""
    quality = np.array([0, 0, 1])
    assert np.all(KeplerQualityFlags.create_quality_mask(quality, bitmask=0))
    assert np.all(KeplerQualityFlags.create_quality_mask(quality, bitmask=None))
    assert np.all(KeplerQualityFlags.create_quality_mask(quality, bitmask='none'))
    assert (KeplerQualityFlags.create_quality_mask(quality, bitmask=1)).sum() == 2
    assert (KeplerQualityFlags.create_quality_mask(quality, bitmask='hardest')).sum() == 2
    # Do we see a ValueError if an invalid bitmask is passed?
    with pytest.raises(ValueError) as err:
        KeplerQualityFlags.create_quality_mask(quality, bitmask='invalidoption')
    assert "not supported" in err.value.args[0]


@pytest.mark.xfail  # Lightkurve v2.x no longer support NaNs in time values
def test_lightkurve_warning():
    """Can we ignore Lightkurve warnings?"""
    with warnings.catch_warnings(record=True) as warns:
        warnings.simplefilter('ignore', LightkurveWarning)
        time = np.array([1, 2, 3, np.nan])
        flux = np.array([1, 2, 3, 4])
        lc = LightCurve(time=time, flux=flux)
        assert len(warns) == 0


def test_validate_method():
    assert validate_method("foo", ["foo", "bar"]) == "foo"
    assert validate_method("FOO", ["foo", "bar"]) == "foo"
    with pytest.raises(ValueError):
          validate_method("foo", ["bar"])


def test_import():
    """Regression test for #605; `lk.utils` resolved to `lk.seismology.utils`"""
    from .. import utils
    assert hasattr(utils, "btjd_to_astropy_time")


def test_btjd_bkjd_input():
    """Regression test for #607: are the bkjd/btjd functions tolerant?"""
    # Kepler
    assert bkjd_to_astropy_time(0).jd[0] == 2454833.
    for user_input in [[0], np.array([0])]:
        assert_array_equal(bkjd_to_astropy_time(user_input).jd, np.array([2454833.]))
    # TESS
    assert btjd_to_astropy_time(0).jd[0] == 2457000.
    for user_input in [[0], np.array([0])]:
        assert_array_equal(btjd_to_astropy_time(user_input).jd, np.array([2457000.]))


def test_centroid_quadratic():
    """Test basic operation of the quadratic centroiding function."""
    # Single bright pixel in the center
    data = np.ones((9, 9))
    data[2, 5] = 10
    col, row = centroid_quadratic(data)
    assert np.isclose(row, 2) & np.isclose(col, 5)

    # Two equally-bright pixels side by side
    data = np.zeros((9, 9))
    data[5, 1] = 5
    data[5, 2] = 5
    col, row = centroid_quadratic(data)
    assert np.isclose(row, 5) & np.isclose(col, 1.5)


def test_centroid_quadratic_robustness():
    """Test quadratic centroids in edge cases; regression test for #610."""
    # Brightest pixel in upper left
    data = np.zeros((5, 5))
    data[0, 0] = 1
    centroid_quadratic(data)

    # Brightest pixel in bottom right
    data = np.zeros((5, 5))
    data[-1, -1] = 1
    centroid_quadratic(data)

    # Data contains a NaN
    data = np.zeros((5, 5))
    data[0, 0] = np.nan
    data[-1, -1] = 10
    col, row = centroid_quadratic(data)
    assert np.isfinite(col) & np.isfinite(row)


def test_show_citation_instructions():
    show_citation_instructions()



import pytest

from astropy.io import fits
import numpy as np
from numpy.testing import assert_array_equal

from ..k2sff import read_k2sff_lightcurve
from ... import search_lightcurve


@pytest.mark.remote_data
def test_read_k2sff():
    """Can we read K2SFF files?"""
    url = "http://archive.stsci.edu/hlsps/k2sff/c16/212100000/00236/hlsp_k2sff_k2_lightcurve_212100236-c16_kepler_v1_llc.fits"
    f = fits.open(url)
    # Verify different extensions
    fluxes = []
    for ext in ["BESTAPER", "CIRC_APER9"]:
        lc = read_k2sff_lightcurve(url, ext=ext)
        assert type(lc).__name__ == "KeplerLightCurve"
        # Are `time` and `flux` consistent with the FITS file?
        assert_array_equal(f[ext].data['T'], lc.time.value)
        assert_array_equal(f[ext].data['FCOR'], lc.flux.value)
        fluxes.append(lc.flux)
    # Different extensions should show different fluxes
    assert not np.array_equal(fluxes[0] , fluxes[1])


@pytest.mark.remote_data
def test_search_k2sff():
    """Can we search and download a K2SFF light curve?"""
    # Try an early campaign
    search = search_lightcurve("K2-18", author="K2SFF", campaign=1)
    assert len(search) == 1
    assert search.table["author"][0] == "K2SFF"
    lc = search.download()
    assert type(lc).__name__ == "KeplerLightCurve"
    assert lc.campaign == 1
    # Try a late campaign
    lc = search_lightcurve("GJ 9827", author="K2SFF", campaign=19).download()
    assert type(lc).__name__ == "KeplerLightCurve"
    assert lc.targetid == 246389858
    assert lc.campaign == 19



"""Tests the `lightkurve.correctors.SFFCorrector` class."""
import pytest
import warnings

import numpy as np
from astropy.utils.data import get_pkg_data_filename
from numpy.testing import assert_array_equal

from ... import LightCurve, KeplerLightCurve, \
                TessLightCurve, LightkurveWarning, \
                search_lightcurve
from .. import SFFCorrector


K2_C08 = ("https://archive.stsci.edu/missions/k2/lightcurves/c8/"
          "220100000/39000/ktwo220139473-c08_llc.fits")


@pytest.mark.remote_data
@pytest.mark.parametrize("path", [K2_C08])
def test_remote_data(path):
    """Can we correct a simple K2 light curve?"""
    lc = KeplerLightCurve.read(path, quality_bitmask=None)
    sff = SFFCorrector(lc.remove_nans())
    sff.correct(windows=10, bins=5, timescale=0.5)
    sff.correct(windows=10, bins=5, timescale=0.5, sparse=True)


def test_sff_knots():
    """Is SFF robust against gaps in time and irregular time sampling?
    This test creates a light curve with gaps in time between
    days 20-30 and days 78-80.  In addition, the time sampling rate changes
    in the interval between day 30 and 78.  SFF should fail without error.
    """
    n_points = 300
    fn = get_pkg_data_filename('../../tests/data/ep60021426alldiagnostics.csv')
    data = np.genfromtxt(fn, delimiter=',', skip_header=1)
    raw_flux = data[:, 1][:n_points]
    centroid_col = data[:, 3][:n_points]
    centroid_row = data[:, 4][:n_points]

    time = np.concatenate((np.linspace(0, 20, int(n_points/3)),
                           np.linspace(30, 78, int(n_points/3)),
                           np.linspace(80, 100, int(n_points/3))
                           ))
    lc = KeplerLightCurve(time=time,
                          flux=raw_flux,
                          flux_err=np.ones(n_points) * 0.0001,
                          centroid_col=centroid_col,
                          centroid_row=centroid_row)

    # These calls should not raise an exception:
    SFFCorrector(lc).correct()
    lc.to_corrector(method="sff").correct()


def test_sff_corrector():
    """Does our code agree with the example presented in Vanderburg
    and Johnson (2014)?"""
    # The following csv file, provided by Vanderburg and Johnson
    # at https://www.cfa.harvard.edu/~avanderb/k2/ep60021426.html,
    # contains the results of applying SFF to EPIC 60021426.
    fn = get_pkg_data_filename('../../tests/data/ep60021426alldiagnostics.csv')
    data = np.genfromtxt(fn, delimiter=',', skip_header=1)
    mask = data[:, -2] == 0  # indicates whether the thrusters were on or off
    time = data[:, 0]
    raw_flux = data[:, 1]
    corrected_flux = data[:, 2]
    centroid_col = data[:, 3]
    centroid_row = data[:, 4]

    # NOTE: we need a small number of windows below because this test data set
    # is unusually short, i.e. has an unusually small number of cadences.
    lc = LightCurve(time=time, flux=raw_flux, flux_err=np.ones(len(raw_flux)) * 0.0001)
    sff = SFFCorrector(lc)
    corrected_lc = sff.correct(centroid_col=centroid_col,
                               centroid_row=centroid_row,
                               restore_trend=True,
                               windows=1)
    assert (np.isclose(corrected_flux, corrected_lc.flux, atol=0.001).all())
    assert len(sff.window_points) == 0  # expect 0 break points for 1 window

    # masking
    corrected_lc = sff.correct(centroid_col=centroid_col,
                               centroid_row=centroid_row,
                               windows=3,
                               restore_trend=True,
                               cadence_mask=mask)
    assert (np.isclose(corrected_flux, corrected_lc.flux, atol=0.001).all())
    assert len(sff.window_points) == 2  # expect 2 break points for 3 windows

    # masking and breakindex
    corrected_lc = sff.correct(centroid_col=centroid_col,
                               centroid_row=centroid_row,
                               windows=3,
                               restore_trend=True,
                               cadence_mask=mask)
    assert (np.isclose(corrected_flux, corrected_lc.flux, atol=0.001).all())

    # masking and breakindex and iters
    corrected_lc = sff.correct(centroid_col=centroid_col,
                               centroid_row=centroid_row, windows=3, restore_trend=True,
                               cadence_mask=mask, niters=3)
    assert (np.isclose(corrected_flux, corrected_lc.flux, atol=0.001).all())

    # masking and breakindex and bins
    corrected_lc = sff.correct(centroid_col=centroid_col,
                               centroid_row=centroid_row, windows=3, restore_trend=True,
                               cadence_mask=mask, bins=5)
    assert (np.isclose(corrected_flux, corrected_lc.flux, atol=0.001).all())
    assert np.all((sff.lc.flux_err/sff.corrected_lc.flux_err) == 1)


    # masking and breakindex and bins and propagate_errors
    corrected_lc = sff.correct(centroid_col=centroid_col,
                               centroid_row=centroid_row, windows=3, restore_trend=True,
                               cadence_mask=mask, bins=5, propagate_errors=True)
    assert (np.isclose(corrected_flux, corrected_lc.flux, atol=0.001).all())
    assert np.all((sff.lc.flux_err/sff.corrected_lc.flux_err) < 1)

    # test using KeplerLightCurve interface
    klc = KeplerLightCurve(time=time,
                           flux=raw_flux,
                           flux_err=np.ones(len(raw_flux)) * 0.0001,
                           centroid_col=centroid_col,
                           centroid_row=centroid_row)
    sff = klc.to_corrector("sff")
    klc = sff.correct(windows=3, restore_trend=True)
    assert (np.isclose(corrected_flux, klc.flux, atol=0.001).all())

    # Can plot
    sff.diagnose()


def test_sff_priors():
    """SFF Spline flux mean should == lc.flux.mean()
    SFF arclength component should have mean 0
    """
    n_points = 300
    fn = get_pkg_data_filename('../../tests/data/ep60021426alldiagnostics.csv')
    data = np.genfromtxt(fn, delimiter=',', skip_header=1)
    raw_flux = data[:, 1][:n_points]
    centroid_col = data[:, 3][:n_points]
    centroid_row = data[:, 4][:n_points]

    time = np.concatenate((np.linspace(0, 20, int(n_points/3)),
                           np.linspace(30, 78, int(n_points/3)),
                           np.linspace(80, 100, int(n_points/3))
                           ))
    lc = KeplerLightCurve(time=time,
                          flux=raw_flux,
                          flux_err=np.ones(n_points) * 0.0001,
                          centroid_col=centroid_col,
                          centroid_row=centroid_row)

    sff = SFFCorrector(lc)
    sff.correct()  # should not raise an exception
    assert np.isclose(sff.diagnostic_lightcurves['spline'].flux.mean(), 1, atol=1e-3)
    assert np.isclose(sff.diagnostic_lightcurves['sff'].flux.mean(), 0, atol=1e-3)


def test_sff_breakindex():
    """Regression test for #616."""
    lc = LightCurve(flux=np.ones(20))
    with warnings.catch_warnings():
        # Ignore "LightkurveWarning: The design matrix has low rank".
        warnings.simplefilter("ignore", LightkurveWarning)
        corr = SFFCorrector(lc)
        corr.correct(breakindex=[5, 10],
                    centroid_col=np.random.randn(20),
                    centroid_row=np.random.randn(20))
        assert 5 in corr.window_points
        assert 10 in corr.window_points
        corr.correct(breakindex=[5, 10],
                    centroid_col=np.random.randn(20),
                    centroid_row=np.random.randn(20), windows=1)
        assert_array_equal(corr.window_points, np.asarray([5, 10]))


def test_sff_tess_warning():
    """SFF is not designed for TESS, so we raise a warning."""
    lc = TessLightCurve(flux=[1, 2, 3], meta={'MISSION': 'TESS'})
    with pytest.warns(LightkurveWarning, match='not suitable'):
        corr = SFFCorrector(lc)


@pytest.mark.remote_data
def test_sff_nan_centroids():
    """Regression test for #827: SFF failed if light curve contained
    NaNs in its `centroid_col` or `centroid_row` columns."""
    lc = search_lightcurve("EPIC 211083408").download()
    # This previously raised a ValueError:
    lc[200:500].remove_nans().to_corrector("sff").correct()



"""The .io sub-package provides functions for reading data."""
from .detect import *
from .read import *

from . import kepler, tess, qlp, k2sff, everest
from .. import LightCurve

from astropy.io import registry


__all__ = ['read', 'open']


# We intend the reader functions to be accessed via `LightCurve.read()`,
# so we add them to the `astropy.io.registry`.
try:
    registry.register_reader('kepler', LightCurve, kepler.read_kepler_lightcurve)
    registry.register_reader('tess', LightCurve, tess.read_tess_lightcurve)
    registry.register_reader('qlp', LightCurve, qlp.read_qlp_lightcurve)
    registry.register_reader('k2sff', LightCurve, k2sff.read_k2sff_lightcurve)
    registry.register_reader('everest', LightCurve, everest.read_everest_lightcurve)
except registry.IORegistryError:
    pass  # necessary to enable autoreload during debugging
